"""Implementation of the scoreboard command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.je.types import RenderType


__all__ = [
    'ModifyProxy',
    'ObjectivesProxy',
    'PlayersProxy',
    'ScoreboardProxy',
    'scoreboard'
]


class ModifyProxy(CommandProxy):
    """Proxy for modify commands."""

    def displayname(self, objective: str, display_name: str) -> str:
        """Modifies the display name of the objective."""
        return self._run(objective, 'displayname', display_name)

    def rendertype(self, objective: str, render_type: RenderType):
        """Modifies the render type of the objective."""
        return self._run(objective, 'rendertype', render_type)


class ObjectivesProxy(CommandProxy):
    """Proxy for objectives sub commands."""

    def add(
            self,
            objective: str,
            criterion: str,
            display_name: Optional[str] = None
    ) -> str:
        """Adds an objective to the scoreboard."""
        return self._run('add', objective, criterion, display_name)

    def list(self) -> str:
        """Lists objectives on the scoreboard."""
        return self._run('list')

    def modify(self, objective: str) -> ModifyProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.scoreboard.ModifyProxy`
        """
        return self._proxy(ModifyProxy, 'modify', objective)

    def remove(self, objective: str) -> str:
        """Removes an objective from the scoreboard."""
        return self._run('remove', objective)

    def setdisplay(self, slot: str, objective: Optional[str] = None) -> str:
        """Sets the display slot of an objective."""
        return self._run('setdisplay', slot, objective)


class PlayersProxy(CommandProxy):
    """Proxy for player commands."""

    def add(self, targets: str, objective: str, score: int) -> str:
        """Adds score for an objective for a player."""
        return self._run('add', targets, objective, score)

    def enable(self, targets: str, objective: str) -> str:
        """Enables an objective for the given targets."""
        return self._run('enable', targets, objective)

    def get(self, target: str, objective: str) -> str:
        """Gets a player's objective."""
        return self._run('get', target, objective)

    def list(self, target: str = None) -> str:
        """Lists scoreboard of an optional player."""
        return self._run('list', target)

    # pylint: disable=R0913
    def operation(self, targets: str, target_objective: str, operation: str,
                  source: str, source_objective: str) -> str:
        """Applies an arithmetic operation altering the target's/targets'
        score(s) in the target objective, using source target's/targets'
        score(s) in the source objective as input.
        """
        return self._run('operation', targets, target_objective, operation,
                         source, source_objective)

    def remove(self, targets: str, objective: str, score: int) -> str:
        """Decrements the target's/targets' score(s)
        in that objective by the given amount.
        """
        return self._run('remove', targets, objective, score)

    def reset(self, targets: str, objective: Optional[str] = None) -> str:
        """Resets a player's score."""
        return self._run('reset', targets, objective)

    def set(self, targets: str, objective: str, score: int) -> str:
        """Sets a player's score."""
        return self._run('set', targets, objective, score)


class ScoreboardProxy(CommandProxy):
    """Proxy for scoreboard sub commands."""

    @property
    def objectives(self) -> ObjectivesProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.scoreboard.ObjectivesProxy`
        """
        return self._proxy(ObjectivesProxy, 'objectives')

    @property
    def players(self) -> PlayersProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.scoreboard.PlayersProxy`
        """
        return self._proxy(PlayersProxy, 'players')


def scoreboard(self: Client) -> ScoreboardProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.scoreboard.ScoreboardProxy`
    """

    return ScoreboardProxy(self, 'scoreboard')



"""Handshake protocol."""

from __future__ import annotations
from contextlib import suppress
from typing import IO, NamedTuple

from mcipc.query.proto.common import MAGIC
from mcipc.query.proto.common import NULL
from mcipc.query.proto.common import random_session_id
from mcipc.query.proto.common import BigEndianSignedInt32
from mcipc.query.proto.common import Type


__all__ = ['Request', 'Response']


class Request(NamedTuple):
    """A client → server handshake request packet."""

    magic: bytes = MAGIC
    type: Type = Type.HANDSHAKE
    session_id: BigEndianSignedInt32 = BigEndianSignedInt32()

    def __bytes__(self):
        """Converts the packet to bytes."""
        payload = self.magic
        payload += bytes(self.type)
        payload += bytes(self.session_id)
        return payload

    @classmethod
    def create(cls) -> Request:
        """Creates a handshake with a random session ID."""
        return cls(session_id=random_session_id())


class Response(NamedTuple):
    """A server → client handshake response packet."""

    type: bytes
    session_id: BigEndianSignedInt32
    challenge_token: BigEndianSignedInt32

    @classmethod
    def read(cls, file: IO) -> Response:
        """Reads the response from a file-like object."""
        type_ = Type.read(file)
        session_id = BigEndianSignedInt32.read(file)
        # For challenge token, see: https://wiki.vg/Query#Handshake
        bytes_ = b''

        while True:
            byte = file.read(1)

            if byte == NULL:
                with suppress(ValueError):
                    challenge_token = BigEndianSignedInt32(bytes_.decode())
                    break

            bytes_ += byte

        return cls(type_, session_id, challenge_token)

    def to_json(self) -> dict:
        """Returns a JSON-ish dict."""
        return {
            'type': self.type.value,
            'session_id': self.session_id,
            'challenge_token': self.challenge_token
        }



"""Client implementation for Bedrock Edition."""

from mcipc.rcon.be.commands.ability import ability
from mcipc.rcon.be.commands.camerashake import camerashake
from mcipc.rcon.be.commands.clear import clear
from mcipc.rcon.be.commands.clearspawnpoint import clearspawnpoint
from mcipc.rcon.be.commands.clone import clone
from mcipc.rcon.be.commands.difficulty import difficulty
from mcipc.rcon.be.commands.effect import effect
from mcipc.rcon.be.commands.enchant import enchant
from mcipc.rcon.be.commands.event import event
from mcipc.rcon.be.commands.fill import fill
from mcipc.rcon.be.commands.fog import fog
from mcipc.rcon.be.commands.gamemode import gamemode
from mcipc.rcon.be.commands.give import give
from mcipc.rcon.be.commands.help import help    # pylint: disable=W0622
from mcipc.rcon.be.commands.list import list    # pylint: disable=W0622
from mcipc.rcon.be.commands.locate import locate
from mcipc.rcon.be.commands.mixer import mixer
from mcipc.rcon.be.commands.mobevent import mobevent
from mcipc.rcon.be.commands.music import music
from mcipc.rcon.be.commands.particle import particle
from mcipc.rcon.be.commands.playanimation import playanimation
from mcipc.rcon.be.commands.playsound import playsound
from mcipc.rcon.be.commands.replaceitem import replaceitem
from mcipc.rcon.be.commands.ride import ride
from mcipc.rcon.be.commands.save import save
from mcipc.rcon.be.commands.schedule import schedule
from mcipc.rcon.be.commands.scoreboard import scoreboard
from mcipc.rcon.be.commands.setblock import setblock
from mcipc.rcon.be.commands.setmaxplayers import setmaxplayers
from mcipc.rcon.be.commands.setworldspawn import setworldspawn
from mcipc.rcon.be.commands.spawnpoint import spawnpoint
from mcipc.rcon.be.commands.spreadplayers import spreadplayers
from mcipc.rcon.be.commands.stopsound import stopsound
from mcipc.rcon.be.commands.structure import structure
from mcipc.rcon.be.commands.summon import summon
from mcipc.rcon.be.commands.teleport import teleport
from mcipc.rcon.be.commands.testfor import testfor
from mcipc.rcon.be.commands.testforblock import testforblock
from mcipc.rcon.be.commands.testforblocks import testforblocks
from mcipc.rcon.be.commands.tickingarea import tickingarea
from mcipc.rcon.be.commands.time import time
from mcipc.rcon.be.commands.title import title
from mcipc.rcon.be.commands.toggledownfall import toggledownfall
from mcipc.rcon.be.commands.transferserver import transferserver
from mcipc.rcon.be.commands.weather import weather
from mcipc.rcon.be.commands.worldbuilder import wb, worldbuilder
from mcipc.rcon.be.commands.wsserver import connect, wsserver
from mcipc.rcon.be.commands.xp import xp
from mcipc.rcon.commands.chat import me, say, tell, tellraw, send_url
from mcipc.rcon.commands.execute import execute
from mcipc.rcon.commands.function import function
from mcipc.rcon.commands.gamerule import gamerule
from mcipc.rcon.commands.kick import kick
from mcipc.rcon.commands.kill import kill
from mcipc.rcon.commands.op import deop, op
from mcipc.rcon.commands.reload import reload
from mcipc.rcon.commands.tag import tag
from mcipc.rcon.commands.whitelist import whitelist
from mcipc.rcon.client import Client as _Client


__all__ = ['Client']


class Client(_Client):
    """RCON client for the Bedrock Edition."""

    ability = ability
    camerashake = property(camerashake)
    clear = clear
    clearspawnpoint = clearspawnpoint
    clone = clone
    connect = connect
    deop = deop
    difficulty = difficulty
    effect = effect
    enchant = enchant
    event = property(event)
    execute = property(execute)
    fill = fill
    fog = fog
    function = function
    gamemode = gamemode
    gamerule = gamerule
    give = give
    help = help
    kick = kick
    kill = kill
    list = list
    locate = locate
    me = me
    mixer = mixer
    mobevent = mobevent
    music = property(music)
    op = op
    particle = particle
    playanimation = playanimation
    playsound = playsound
    reload = reload
    replaceitem = property(replaceitem)
    ride = ride
    save = save
    say = say
    schedule = property(schedule)
    scoreboard = property(scoreboard)
    send_url = send_url
    setblock = setblock
    setmaxplayers = setmaxplayers
    setworldspawn = setworldspawn
    spawnpoint = spawnpoint
    spreadplayers = spreadplayers
    stopsound = stopsound
    structure = property(structure)
    summon = summon
    tag = tag
    teleport = teleport
    tell = msg = w = tell
    tellraw = tellraw
    testfor = testfor
    testforblock = testforblock
    testforblocks = testforblocks
    tickingarea = property(tickingarea)
    time = property(time)
    title = title
    toggledownfall = toggledownfall
    tp = teleport
    transferserver = transferserver
    wb = wb
    weather = property(weather)
    whitelist = property(whitelist)
    worldbuilder = worldbuilder
    wsserver = wsserver
    xp = experience = xp



"""Implementation of the recipe command."""

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy


__all__ = ['RecipeProxy', 'recipe']


class RecipeProxy(CommandProxy):
    """Proxy for recipe related commands."""

    # pylint: disable=W0621
    def give(self, targets: str, recipe: str = '*') -> str:
        """Gives or takes all recipes from that player."""
        return self._run('give', targets, recipe)

    def take(self, targets: str, recipe: str = '*') -> str:
        """Gives or takes a specified recipe from that player."""
        return self._run('take', targets, recipe)


def recipe(self: Client) -> RecipeProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.recipe.RecipeProxy`
    """

    return RecipeProxy(self, 'recipe')



"""Response types."""

from mcipc.rcon.response_types.kick import KickedPlayer
from mcipc.rcon.response_types.location import Location
from mcipc.rcon.response_types.players import Player, Players


__all__ = ['KickedPlayer', 'Location', 'Player', 'Players']



"""Implementation of the setmaxplayers command."""

from mcipc.rcon.client import Client


__all__ = ['setmaxplayers']


def setmaxplayers(self: Client, max_players: int) -> str:
    """Sets the maximally allowed players on the server."""

    return self.run('setmaxplayers', max_players)



"""Implementation of the attribute command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import Action


__all__ = [
    'AttributeProxy',
    'BaseProxy',
    'ModifierProxy',
    'ValueProxy',
    'attribute'
]


class ValueProxy(CommandProxy):    # pylint: disable=R0903
    """Proxy for attribute value getter."""

    def get(self, uuid: str, scale: Optional[float] = None) -> str:
        """Returns the value of the modifier with the specified UUID."""
        return self._run('get', uuid, scale)


class BaseProxy(CommandProxy):
    """Proxy to modify bases."""

    def get(self, scale: Optional[float] = None) -> str:
        """Returns the base value of the specified attribute."""
        return self._run('get', scale)

    def set(self, value: str) -> str:
        """Overwrites the base value of the specified
        attribute with the given value.
        """
        return self._run('set', value)


class ModifierProxy(CommandProxy):
    """Proxy to modify attributes."""

    def add(self, uuid: str, name: str, value: str, action: Action) -> str:
        """Adds an attribute modifier with the specified properties
        if no modifier with the same UUID already existed.
        """
        return self._run('add', uuid, name, value, action)

    def remove(self, uuid: str) -> str:
        """Removes the attribute modifier with the specified UUID."""
        return self._run('remove', uuid)

    @property
    def value(self) -> ValueProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.attribute.ValueProxy`
        """
        return self._proxy(ValueProxy, 'value')


class AttributeProxy(CommandProxy):
    """Proxy for attribute actions."""

    @property
    def base(self) -> BaseProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.attribute.BaseProxy`
        """
        return self._proxy(BaseProxy, 'base')

    @property
    def modifier(self) -> ModifierProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.attribute.ModifierProxy`
        """
        return self._proxy(ModifierProxy, 'modifier')

    def get(self, scale: Optional[float] = None) -> str:
        """Returns the total value of the specified attribute."""
        return self._run('get', scale)


# pylint: disable=W0621
def attribute(self: Client, target: str, attribute: str) -> AttributeProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.attribute.AttributeProxy`
    """

    return AttributeProxy(self, 'attribute', target, attribute)



"""Implementation of the fill command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.types import FillMode, Vec3


__all__ = ['fill']


# pylint: disable=C0103,R0913
def fill(
        self: Client,
        from_: Vec3,
        to: Vec3,
        block: str,
        tile_data: Optional[int] = None,
        old_block_handling: Optional[FillMode] = None,
        replace_tile_name: Optional[str] = None,
        replace_data_value: Optional[str] = None
) -> str:
    """Fills all or parts of a region with a specific block."""

    return self.run('fill', from_, to, block, tile_data, old_block_handling,
                    replace_tile_name, replace_data_value)



"""Implementation of the help command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.functions import parsed
from mcipc.rcon.response_types.help import parse


__all__ = ['help']


@parsed(parse)
def help(client: Client, command: Optional[str] = None) -> str:
    """Returns help about commands."""

    return client.run('help', command)



"""Implementation of the whitelist command."""

from functools import partial
from re import fullmatch

from mcipc.rcon.client import Client
from mcipc.rcon.functions import parse_bool, parsed
from mcipc.rcon.proxy import CommandProxy


__all__ = ['WhitelistProxy', 'whitelist']


ADDED = r'Added (.+) to the whitelist'
TURNED_OFF = r'Whitelist is now turned off'
TURNED_ON = r'Whitelist is now turned on'
RELOADED = r'Reloaded the whitelist'
REMOVED = r'Removed (.+) from the whitelist'
LIST = r'There are (\d+) whitelisted players: (.*)'


def parse_list(response: str) -> list[str]:
    """Returns a list of whitelisted players."""

    if match := fullmatch(LIST, response):
        return list(filter(None, map(str.strip, match.group(2).split(','))))

    return []


class WhitelistProxy(CommandProxy):
    """Proxy for whitelist commands."""

    @parsed(partial(parse_bool, true=ADDED, default=False))
    def add(self, name: str) -> str:
        """Adds a player to the whitelist."""
        return self._run('add', name)

    @parsed(parse_list)
    def list(self) -> str:
        """Lists the whitelist."""
        return self._run('list')

    @parsed(partial(parse_bool, true=TURNED_OFF, default=False))
    def off(self) -> str:
        """Turns the whitelist checking off."""
        return self._run('off')

    @parsed(partial(parse_bool, true=TURNED_ON, default=False))
    def on(self) -> str:    # pylint: disable=C0103
        """Turns the whitelist checking on."""
        return self._run('on')

    @parsed(partial(parse_bool, true=RELOADED, default=False))
    def reload(self) -> str:
        """Reloads the whitelist from the file."""
        return self._run('reload')

    @parsed(partial(parse_bool, true=REMOVED, default=False))
    def remove(self, name: str) -> str:
        """Removes a player from the whitelist."""
        return self._run('remove', name)


def whitelist(self: Client) -> WhitelistProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.commands.whitelist.WhitelistProxy`
    """

    return WhitelistProxy(self, 'whitelist')



"""Implementation of the playsound command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.types import Vec3


__all__ = ['playsound']


# pylint: disable=R0913
def playsound(
        self: Client,
        sound: str,
        player: Optional[str] = None,
        position: Optional[Vec3] = None,
        volume: Optional[float] = None,
        minimum_volume: Optional[float] = None
) -> str:
    """Plays a sound."""

    return self.run('playsound', sound, player, position, volume,
                    minimum_volume)



"""Implementation of the event command."""

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy


__all__ = ['EventProxy', 'event']


class EventProxy(CommandProxy):     # pylint: disable=R0903
    """Proxies event-related commands."""

    def entity(self, target: str, event_name: str) -> str:
        """Triggers an event on an entity."""
        return self._run('entity', target, event_name)


def event(self: Client) -> EventProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.be.commands.event.EventProxy`
    """

    return EventProxy(self, 'event')



"""Implementation of the execute command."""

from __future__ import annotations
from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import Anchor
from mcipc.rcon.types import BossbarSlot
from mcipc.rcon.types import DataType
from mcipc.rcon.types import IntRange
from mcipc.rcon.types import Rotation
from mcipc.rcon.types import ScanMode
from mcipc.rcon.types import StorageType
from mcipc.rcon.types import Vec3


__all__ = [
    'ConditionalProxy',
    'DataProxy',
    'ExecuteProxy',
    'ScoreProxy',
    'StoreProxy',
    'execute'
]


class ScoreProxy(CommandProxy):
    """Check a score against either another score or a given range."""

    # pylint: disable=W0622
    def matches(self, range: IntRange) -> ExecuteProxy:
        """Matches a range."""
        return self._proxy(ExecuteProxy, 'matches', range)

    # pylint: disable=C0103
    def lt(self, source: str, objective: str) -> ExecuteProxy:
        """Returns a < comparison."""
        return self._proxy(ExecuteProxy, '<', source, objective)

    def le(self, source: str, objective: str) -> ExecuteProxy:
        """Returns a <= comparison."""
        return self._proxy(ExecuteProxy, '<=', source, objective)

    def eq(self, source: str, objective: str) -> ExecuteProxy:
        """Returns a == comparison."""
        return self._proxy(ExecuteProxy, '==', source, objective)

    def ge(self, source: str, objective: str) -> ExecuteProxy:
        """Returns a >= comparison."""
        return self._proxy(ExecuteProxy, '>=', source, objective)

    def gt(self, source: str, objective: str) -> ExecuteProxy:
        """Returns a > comparison."""
        return self._proxy(ExecuteProxy, '>', source, objective)


class DataProxy(CommandProxy):
    """Checks whether the targeted block, entity
    or storage has any data for a given tag.
    """

    def block(self, pos: Vec3, path: str) -> ExecuteProxy:
        """Checks a block."""
        return self._proxy(ExecuteProxy, 'block', pos, path)

    def entity(self, target: str, path: str) -> ExecuteProxy:
        """Checks an entity."""
        return self._proxy(ExecuteProxy, 'entity', target, path)

    def storage(self, source: str, path: str) -> ExecuteProxy:
        """Checks a storage."""
        return self._proxy(ExecuteProxy, 'storage', source, path)


class ConditionalProxy(CommandProxy):
    """Proxies if and unless conditionals."""

    def block(self, pos: Vec3, block: str) -> ExecuteProxy:
        """Compares the block at a given position to a given block ID."""
        return self._proxy(ExecuteProxy, 'block', pos, block)

    def blocks(self, start: Vec3, end: Vec3, destination: Vec3,
               scan_mode: ScanMode) -> ExecuteProxy:
        """Compares the blocks in two equally sized volumes."""
        return self._proxy(ExecuteProxy, 'blocks', start, end, destination,
                           scan_mode)

    @property
    def data(self) -> DataProxy:
        """Delegates to a :py:class:`mcipc.rcon.commands.execute.DataProxy`"""
        return self._proxy(DataProxy, 'data')

    def entity(self, entities: str) -> ExecuteProxy:
        """Checks whether one or more <targets> exist."""
        return self._proxy(ExecuteProxy, 'entity', entities)

    def predicate(self, predicate: str) -> ExecuteProxy:
        """Checks whether the <predicate> evaluates to a positive result."""
        return self._proxy(ExecuteProxy, 'predicate', predicate)

    def score(self, target: str, target_objective: str) -> ScoreProxy:
        """Delegates to a score proxy."""
        return self._proxy(ScoreProxy, 'score', target, target_objective)


class StoreProxy(CommandProxy):
    """Proxy to handle store sub-command."""

    def block(
            self,
            target_pos: Vec3,
            path: str,
            type: DataType,
            scale: float
    ) -> ExecuteProxy:
        """Saves the final command's return value
        as tag data within a block entity.
        """
        return self._proxy(ExecuteProxy, 'block', target_pos, path, type,
                           scale)

    def bossbar(self, ident: str, value: BossbarSlot) -> ExecuteProxy:
        """Saves the final command's return value in either
        a bossbar's current value or its maximum value.
        """
        return self._proxy(ExecuteProxy, 'bossbar', ident, value)

    def entity(
            self,
            target: str,
            path: str,
            type: DataType,
            scale: float
    ) -> ExecuteProxy:
        """Save the final command's return value
        in one of an entity's data tags.
        """
        return self._proxy(ExecuteProxy, 'entity', target, path, type, scale)

    def score(self, targets: str, objective: str) -> ExecuteProxy:
        """Overrides the score held by <targets> on the given
        <objective> with the final command's return value.
        """
        return self._proxy(ExecuteProxy, 'score', targets, objective)

    def storage(
            self,
            target: str,
            path: str,
            type: DataType,
            scale: float
    ) -> ExecuteProxy:
        """Uses the <path> within storage <target>
        to store the return value in.
        """
        return self._proxy(ExecuteProxy, 'storage', target, path, type, scale)


class ExecuteProxy(CommandProxy):
    """A proxy for execute sub-commands."""

    def __call__(self) -> str:
        """Runs the command in the current execution context."""
        return self._run()

    def align(self, axes: str) -> ExecuteProxy:
        """Aligns with the given axes and returns a sub-proxy."""
        return self._proxy(type(self), 'align', axes)

    def anchored(self, anchor: str) -> ExecuteProxy:
        """Stores the distance from the feet to the eyes of
        the entity that is executing the command in the anchor,
        which is part of the command context.
        """
        return self._proxy(type(self), 'anchored', anchor)

    def as_(self, targets: str) -> ExecuteProxy:
        """Executes the command as the given targets."""
        return self._proxy(type(self), 'as', targets)

    def at(self, targets: str) -> ExecuteProxy:     # pylint: disable=C0103
        """Executes the command at the given targets."""
        return self._proxy(type(self), 'at', targets)

    def facing(
            self,
            *,
            pos: Optional[Vec3] = None,
            targets: Optional[str] = None,
            anchor: Optional[Anchor] = None
    ) -> ExecuteProxy:
        """Faces a position or entity."""
        if sum(item is not None for item in (pos, targets)) != 1:
            raise ValueError('Must specify either pos or targets.')

        if pos is not None:
            return self._proxy(type(self), 'facing', pos)

        if anchor is None:
            raise ValueError('No anchor specified.')

        return self._proxy(type(self), 'facing', 'entity', targets, anchor)

    def in_(self, dimension: str) -> ExecuteProxy:
        """Executes in the given dimension."""
        return self._proxy(type(self), 'in', dimension)

    def positioned(
            self,
            *,
            pos: Optional[Vec3] = None,
            targets: Optional[str] = None
    ) -> ExecuteProxy:
        """Runs the command with a given positioning."""
        if sum(item is not None for item in (pos, targets)) != 1:
            raise ValueError('Must specify either pos or targets.')

        if pos is not None:
            return self._proxy(type(self), 'positioned', pos)

        return self._proxy(type(self), 'positioned', 'as', targets)

    def rotated(
            self,
            *,
            rot: Optional[Rotation] = None,
            targets: Optional[str] = None
    ) -> ExecuteProxy:
        """Executes the command with the given rotation."""
        if sum(item is not None for item in (rot, targets)) != 1:
            raise ValueError('Must specify either rot or targets.')

        if rot is not None:
            return self._proxy(type(self), 'rotated', rot)

        return self._proxy(type(self), 'rotated', 'as', targets)

    def store(self, what: StorageType):
        """Stores the result or success."""
        return self._proxy(StoreProxy, 'store', what)

    @property
    def if_(self) -> ConditionalProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.commands.execute.ConditionalProxy`
        """
        return self._proxy(ConditionalProxy, 'if')

    @property
    def unless(self) -> ConditionalProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.commands.execute.ConditionalProxy`
        """
        return self._proxy(ConditionalProxy, 'unless')

    def run(self, *arguments: str) -> str:
        """Runs the a command with arguments
        in the current execution context.
        """
        return self._run('run', *arguments)


def execute(self: Client) -> ExecuteProxy:
    """Delegates to a :py:class:`mcipc.rcon.commands.execute.ExecuteProxy`"""

    return ExecuteProxy(self, 'execute')



"""Implementation of the playsound command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.je.types import SoundSource
from mcipc.rcon.types import Vec3


__all__ = ['playsound']


# pylint: disable=R0913
def playsound(
        self: Client,
        sound: str,
        source: SoundSource,
        targets: str,
        pos: Optional[Vec3] = None,
        volume: Optional[float] = None,
        pitch: Optional[float] = None,
        min_volume: Optional[float] = None
) -> str:
    """Plays a sound."""

    return self.run(
        'playsound', sound, source, targets, pos, volume, pitch, min_volume
    )



"""Implementation of the wsserver and connect commands."""

from mcipc.rcon.client import Client


__all__ = ['connect', 'wsserver']


def connect(self: Client, server_uri: str) -> str:
    """Connect to designated WebSocket server."""

    return self.run('connect', server_uri)


def wsserver(self: Client, server_uri: str) -> str:
    """Connect to designated WebSocket server."""

    return self.run('wsserver', server_uri)



"""Tests the basic_stats module."""

from ipaddress import IPv4Address
from unittest import TestCase

from mcipc.query.proto.common import Type, random_session_id
from mcipc.query.proto.basic_stats import BasicStats


class TestBasicStats(TestCase):
    """Tests the BasicStats named tuple."""

    def setUp(self):
        """Sets up test data."""
        self.type = Type.STAT
        self.session_id = random_session_id()
        self.ip_address = IPv4Address('127.0.0.1')
        self.basic_stats = BasicStats(
            self.type,
            self.session_id,
            'Message of the day',
            'survival',
            'world',
            1,
            20,
            25565,
            self.ip_address
        )
        self.json = {
            'type': int(self.type.value),
            'session_id': int(self.session_id),
            'motd': 'Message of the day',
            'game_type': 'survival',
            'map': 'world',
            'num_players': 1,
            'max_players': 20,
            'host_port': 25565,
            'host_ip': str(self.ip_address)
        }

    def test_json_serialization(self):
        """Tests JSON serialization of BasticStats."""
        self.assertEqual(dict(self.basic_stats), self.json)



"""Types for client -> Java Edition server communication."""

from typing import Union

from mcipc.rcon.je.enumerations import Biome
from mcipc.rcon.je.enumerations import CollisionRule
from mcipc.rcon.je.enumerations import Color
from mcipc.rcon.je.enumerations import Enchantment
from mcipc.rcon.je.enumerations import ParticleMode
from mcipc.rcon.je.enumerations import RenderType
from mcipc.rcon.je.enumerations import ScheduleMode
from mcipc.rcon.je.enumerations import SoundSource
from mcipc.rcon.je.enumerations import Structure
from mcipc.rcon.je.enumerations import TeamOption
from mcipc.rcon.je.enumerations import TimeSpec
from mcipc.rcon.je.enumerations import TimeUnit
from mcipc.rcon.je.enumerations import Visibility


__all__ = [
    'Biome',
    'CollisionRule',
    'Color',
    'Enchantment',
    'ParticleMode',
    'RenderType',
    'ScheduleMode',
    'SoundSource',
    'Structure',
    'TeamOption',
    'TeamValue',
    'Time',
    'TimeSpec',
    'TimeUnit',
    'Visibility'
]


Biome = Union[Biome, str]
CollisionRule = Union[CollisionRule, str]
Color = Union[Color, str]
Enchantment = Union[Enchantment, str]
ParticleMode = Union[ParticleMode, str]
RenderType = Union[RenderType, str]
ScheduleMode = Union[ScheduleMode, str]
SoundSource = Union[SoundSource, str]
Structure = Union[Structure, str]
TeamOption = Union[TeamOption, str]
Time = Union[str, int, float]
TimeSpec = Union[TimeSpec, str]
TimeUnit = Union[TimeUnit, str]
Visibility = Union[Visibility, str]
TeamValue = Union[CollisionRule, Color, Visibility, dict, bool, str]



"""Implementation of the kick command."""

from mcipc.rcon.client import Client
from mcipc.rcon.functions import parsed
from mcipc.rcon.response_types.kick import parse


__all__ = ['kick']


@parsed(parse)
def kick(self: Client, player: str, *reasons: str) -> str:
    """Kicks the respective player."""

    return self.run('kick', player, *reasons)



"""Implementation of the spreadplayers command."""

from mcipc.rcon.be.types import RelativeFloat
from mcipc.rcon.client import Client


__all__ = ['spreadplayers']


def spreadplayers(
        self: Client,
        x: RelativeFloat,
        y: RelativeFloat,
        spread_distance: float,
        max_range: float,
        victim: str
) -> str:
    """Spreads players."""

    return self.run('spreadplayers', x, y, spread_distance, max_range, victim)



"""Implementation of the forceload command."""

from typing import Optional, Union

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import Vec2


__all__ = ['ForceloadProxy', 'forceload']


class ForceloadProxy(CommandProxy):
    """Provides forceload-related sub commands."""

    # pylint: disable=C0103
    def add(self, from_: Vec2, to: Optional[Vec2] = None) -> str:
        """Forces the chunk at the <from> position (through to <to> if set) in
        the dimension of the command's execution to be loaded constantly.
        """
        return self._run('add', from_, to)

    def remove(
            self,
            from_or_all: Union[Vec2, 'all'],
            to: Optional[Vec2] = None
    ) -> str:
        """Unforces the chunk at the <from> position (through to <to> if set)
        in the dimension of the command's execution to be loaded constantly.
        """
        if from_or_all == 'all':
            return self._run('remove', 'all')

        if to is None:
            raise ValueError('To must be specified, if not removing all.')

        return self._run('remove', from_or_all, to)

    def query(self, pos: Optional[Vec2] = None) -> str:
        """If chunk coordinates are given, displays the specified chunk in the
        dimension of the command's execution is force loaded; otherwise, lists
        which chunks in the dimension of the command's execution are force
        loaded.
        """
        return self._run('query', pos)


def forceload(self: Client) -> ForceloadProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.forceload.ForceloadProxy`
    """

    return ForceloadProxy(self, 'forceload')



"""Implementation of the title comman."""

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy


__all__ = ['TitleProxy', 'title']


class TitleProxy(CommandProxy):
    """Proxy for title commands."""

    def clear(self) -> str:
        """Clears the title."""
        return self._run('clear')

    def reset(self) -> str:
        """Resets the title."""
        return self._run('reset')

    def title(self, title_text: str) -> str:
        """Sets the title."""
        return self._run('title', title_text)

    def subtitle(self, title_text: str) -> str:
        """Sets the subtitle."""
        return self._run('subtitle', title_text)

    def actionbar(self, title_text: str) -> str:
        """Sets the action bar."""
        return self._run('actionbar', title_text)

    def times(self, fade_in: int, stay: int, fade_out: int) -> str:
        """    Changes the fade-in, stay, and fade-out times of all current
        and future screen titles for the specified player(s).
        """
        return self._run('times', fade_in, stay, fade_out)


def title(self: Client, player: str) -> TitleProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.be.commands.title.TitleProxy`
    """

    return TitleProxy(self, 'title', player)



"""Implementation of the give command."""

from typing import Optional

from mcipc.rcon.client import Client


__all__ = ['give']


def give(
        self: Client,
        target: str,
        item: str,
        amount: Optional[int] = None,
        data: Optional[int] = None,
        components: Optional[dict] = None
) -> str:
    """Gives an item to one or more players."""

    return self.run(target, item, amount, data, components)



"""Implementation of the bossbar command."""

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import Attribute, Color, Style


__all__ = ['BossbarProxy', 'BossbarSetProxy', 'bossbar']


class BossbarSetProxy(CommandProxy):
    """Proxy to handle boss bar set commands."""

    def color(self, color: Color) -> str:
        """Set the text color (if no color was specified
        as part of a text component) and bar color.
        Defaults to white upon creation.
        """
        return self._run('color', color)

    def max(self, max: int) -> str:     # pylint: disable=W0622
        """Set the boss bar's maximum value. Defaults to 100 upon creation."""
        return self._run('max', max)

    def name(self, name: str) -> str:   # pylint: disable=W0622
        """Set the boss bar's name."""
        return self._run('name', name)

    def players(self, *targets: str) -> str:
        """Change the set of players to whom the bar is visible.
        Defaults to none upon creation.
        """
        return self._run('players', *targets)

    def style(self, style: Style) -> str:
        """Set the boss bar's visual amount of segments.
        Defaults to progress upon creation.
        """
        return self._run('style', style)

    def value(self, value: int) -> str:
        """Set the boss bar's current value.
        Defaults to 0 upon creation.
        """
        return self._run('value', value)

    def visible(self, visible: bool) -> str:
        """Set the boss bar's visibility.
        Defaults to true upon creation.
        """
        return self._run('visible', visible)


# pylint: disable=W0622,C0103
class BossbarProxy(CommandProxy):
    """Proxy for boss bar methods."""

    def add(self, id: str, name: dict) -> str:
        """Create a new boss bar."""
        return self._run('add', id, name)

    def get(self, id: str, attribute: Attribute) -> str:
        """Return the requested setting as a result of the command."""
        return self._run('get', id, attribute)

    def list(self) -> str:
        """Display a list of existing boss bars."""
        return self._run('list')

    def remove(self, id: str) -> str:
        """Remove an existing bossbar."""
        return self._run('remove', id)

    def set(self, id: str, attribute: Attribute) -> BossbarSetProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.bossbar.BossbarSetProxy`
        """
        return self._proxy(BossbarSetProxy, 'set', id, attribute)


def bossbar(self: Client) -> BossbarProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.bossbar.BossbarProxy`
    """

    return BossbarProxy(self, 'bossbar')



"""Implementation of the replaceitem command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import Vec3


__all__ = ['ReplaceitemProxy', 'replaceitem']


class ReplaceitemProxy(CommandProxy):
    """Proxy for replaceitem related commands."""

    def block(
            self,
            pos: Vec3,
            slot: str,
            item: str,
            count: Optional[int] = None
    ) -> str:
        """Replaces a block."""
        self._run('block', pos, slot, item, count)

    def entity(
            self,
            targets: str,
            slot: str,
            item: str,
            count: Optional[int] = None
    ) -> str:
        """Replaces an entity."""
        self._run('entity', targets, slot, item, count)


def replaceitem(self: Client) -> ReplaceitemProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.replaceitem.ReplaceitemProxy`
    """

    return ReplaceitemProxy(self, 'replaceitem')



"""RCON client library."""

from mcipc.rcon.be import Client as BedrockClient
from mcipc.rcon.ee import Client as EducationClient
from mcipc.rcon.enumerations import Ability
from mcipc.rcon.enumerations import Action
from mcipc.rcon.enumerations import Anchor
from mcipc.rcon.enumerations import Attribute
from mcipc.rcon.enumerations import BossbarSlot
from mcipc.rcon.enumerations import CamerashakeType
from mcipc.rcon.enumerations import CloneMode
from mcipc.rcon.enumerations import Color
from mcipc.rcon.enumerations import DatapackMode
from mcipc.rcon.enumerations import DatapackState
from mcipc.rcon.enumerations import DataType
from mcipc.rcon.enumerations import DebugCommand
from mcipc.rcon.enumerations import Difficulty
from mcipc.rcon.enumerations import Direction
from mcipc.rcon.enumerations import Edition
from mcipc.rcon.enumerations import FillMode
from mcipc.rcon.enumerations import GameMode
from mcipc.rcon.enumerations import Hand
from mcipc.rcon.enumerations import MaskMode
from mcipc.rcon.enumerations import ScanMode
from mcipc.rcon.enumerations import SetblockMode
from mcipc.rcon.enumerations import StorageType
from mcipc.rcon.enumerations import Style
from mcipc.rcon.enumerations import TargetType
from mcipc.rcon.enumerations import TimeType
from mcipc.rcon.enumerations import XPUnit
from mcipc.rcon.exceptions import CommandError
from mcipc.rcon.exceptions import InvalidArgument
from mcipc.rcon.exceptions import InvalidInteger
from mcipc.rcon.exceptions import InvalidNameOrUUID
from mcipc.rcon.exceptions import LocationNotFound
from mcipc.rcon.exceptions import NoPlayerFound
from mcipc.rcon.exceptions import UnexpectedTrailingData
from mcipc.rcon.exceptions import UnknownCommand
from mcipc.rcon.je import Client as JavaClient


__all__ = [
    # Clients
    'CLIENTS',
    'BedrockClient',
    'EducationClient',
    'JavaClient',
    # Exceptions
    'CommandError',
    'InvalidArgument',
    'InvalidInteger',
    'InvalidNameOrUUID',
    'LocationNotFound',
    'NoPlayerFound',
    'UnexpectedTrailingData',
    'UnknownCommand',
    # Enums
    'Ability',
    'Action',
    'Anchor',
    'Attribute',
    'BossbarSlot',
    'CamerashakeType',
    'CloneMode',
    'Color',
    'DatapackMode',
    'DatapackState',
    'DataType',
    'DebugCommand',
    'Difficulty',
    'Direction',
    'Edition',
    'FillMode',
    'GameMode',
    'Hand',
    'MaskMode',
    'ScanMode',
    'SetblockMode',
    'StorageType',
    'Style',
    'TargetType',
    'TimeType',
    'XPUnit'
]


CLIENTS = {
    Edition.BEDROCK: BedrockClient,
    Edition.EDUCATION: EducationClient,
    Edition.JAVA: JavaClient
}
Client = JavaClient     # For backwards compatibility.



"""Implementation of the mobevent command."""

from typing import Optional

from mcipc.rcon.client import Client


__all__ = ['mobevent']


def mobevent(self: Client, event: str, value: Optional[bool] = None) -> str:
    """Controls or queries what mob events are allowed to run."""

    return self.run('mobevent', event, value)



"""Low-level protocol stuff."""

from mcipc.query.proto.basic_stats import BasicStats
from mcipc.query.proto.basic_stats import Request as BasicStatsRequest
from mcipc.query.proto.common import BigEndianSignedInt32
from mcipc.query.proto.full_stats import FullStats
from mcipc.query.proto.full_stats import Request as FullStatsRequest
from mcipc.query.proto.handshake import Request as HandshakeRequest
from mcipc.query.proto.handshake import Response


__all__ = [
    'BasicStats',
    'BasicStatsRequest',
    'BigEndianSignedInt32',
    'FullStats',
    'FullStatsRequest',
    'HandshakeRequest',
    'Response'
]



"""Implementation of the fill command."""

from typing import Optional
from mcipc.rcon.client import Client
from mcipc.rcon.enumerations import FillMode
from mcipc.rcon.types import FillMode as FillModeHint, Vec3


__all__ = ['fill']


def fill(
        self: Client,
        from_: Vec3,
        to: Vec3,
        block: str,
        mode: Optional[FillModeHint] = None,
        filter: Optional[str] = None
) -> str:
    """Fills all or parts of a region with a specific block."""

    command = ['fill', from_, to, block]

    if mode is not None:
        command.append(mode := FillMode(mode))

        if filter is not None and mode == FillMode.REPLACE:
            command.append(filter)

    return self.run(*command)



"""Implementation of the xp command."""

from mcipc.rcon.client import Client


__all__ = ['xp']


def xp(self: Client, amount: int, player: str) -> str:
    """Gives the player the specified amount of XP."""

    return self.run('xp', amount, player)



"""Implementation of the locatebiome command."""

from mcipc.rcon.client import Client
from mcipc.rcon.functions import parsed
from mcipc.rcon.je.types import Biome
from mcipc.rcon.response_types.location import parse


__all__ = ['locatebiome']


@parsed(parse)
def locatebiome(self: Client, biome: Biome) -> str:
    """Locates the given biome."""

    return self.run('locatebiome', biome)



"""Implementation of the music command."""

from typing import Optional

from mcipc.rcon.be.types import MusicRepeatMode
from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy


__all__ = ['MusicProxy', 'music']


class MusicProxy(CommandProxy):
    """Proxy for music commands."""

    def play(
            self,
            track_name: str,
            volume: Optional[float] = None,
            fade_seconds: Optional[float] = None,
            repeat_mode: Optional[MusicRepeatMode] = None
    ) -> str:
        """Plays the music track."""
        return self._run('play', track_name, volume, fade_seconds, repeat_mode)

    def queue(
            self,
            track_name: str,
            volume: Optional[float] = None,
            fade_seconds: Optional[float] = None,
            repeat_mode: Optional[MusicRepeatMode] = None
    ) -> str:
        """Add the music tracks to the queue."""
        return self._run('queue', track_name, volume, fade_seconds,
                         repeat_mode)

    def stop(self, fade_seconds: Optional[float] = None) -> str:
        """Stops the music."""
        return self._run('stop', fade_seconds)

    def volume(self, volume: float) -> str:
        """Adjusting the music volume."""
        return self._run('volume', volume)


def music(self: Client) -> MusicProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.be.commands.music.MusicProxy`
    """

    return MusicProxy(self, 'music')



"""Agent-related commands."""

from mcipc.rcon.client import Client
from mcipc.rcon.types import Direction


__all__ = [
    'collect',
    'createagent',
    'destroy',
    'detect',
    'detectredstone',
    'dropall',
    'move',
    'remove',
    'tpagent',
    'transfer',
    'turn'
]


def collect(self: Client, item: str) -> str:
    """Lets your agent collect all items within
    a one block from Agent in three dimensions.
    """

    return self.run('collect', item)


def createagent(self: Client) -> str:
    """Creates an Agent at the current player's position."""

    return self.run('createagent')


def destroy(self: Client, direction: Direction) -> str:
    """Makes your agent destroy a block or item in the specified direction."""

    return self.run('destroy', direction)


def detect(self: Client, direction: Direction) -> str:
    """Lets your agent detect if there is a
    collidable block in the specified direction.
    """

    return self.run('detect', direction)


def detectredstone(self: Client, direction: Direction) -> str:
    """Lets your agent detect a redstone signal in the specified direction."""

    return self.run('detectredstone', direction)


def dropall(self: Client, direction: Direction) -> str:
    """Makes an agent drop all its items from all slots
    onto the ground by one block in the specified direction.
    """

    return self.run('dropall', direction)


def move(self: Client, direction: Direction) -> str:
    """Moves an Agent in a specified direction."""

    return self.run('move', direction)


def remove(self: Client, target: str) -> str:
    """Removes an agent."""

    return self.run('remove', target)


def tpagent(self: Client) -> str:
    """Teleport the player to their Agent's position."""

    return self.run('tpagent')


def transfer(self: Client, src_slot_num: int, quantity: int,
             dst_slot_num: int) -> str:
    """Transfers a specified quantity of items from
    one slot to another of your Agent's inventory.
    """

    return self.run('transfer', src_slot_num, quantity, dst_slot_num)


def turn(self: Client, direction: Direction) -> str:
    """Rotates an Agent in a specified direction by 90 degrees."""

    return self.run('turn', direction)



"""Result of a successful kick."""

from re import fullmatch
from typing import NamedTuple

from mcipc.functions import json_serializable
from mcipc.rcon.exceptions import NoPlayerFound


__all__ = ['KickedPlayer', 'parse']


REGEX = r'Kicked (.*): (.*)'


@json_serializable
class KickedPlayer(NamedTuple):
    """Stores information about a kicked player."""

    name: str
    reason: str


def parse(text: str) -> KickedPlayer:
    """Parses a kicked player from the text."""

    if (match := fullmatch(REGEX, text)) is not None:
        return KickedPlayer(*match.groups())

    raise NoPlayerFound()



"""Implementation of the camerashake command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import CamerashakeType


__all__ = ['CamerashakeProxy', 'camerashake']


class CamerashakeProxy(CommandProxy):   # pylint: disable=R0903
    """Proxy for camera shake commands."""

    def add(
            self,
            player: str,
            intensity: Optional[float] = None,
            seconds: Optional[float] = None,
            shake_type: Optional[CamerashakeType] = None
    ) -> str:
        """Used to enable a camera shaking effect."""
        return self._run('add', player, intensity, seconds, shake_type)


def camerashake(self: Client) -> CamerashakeProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.be.commands.camerashake.CamerashakeProxy`
    """

    return CamerashakeProxy(self, 'camerashake')



"""Implementation of the kill command."""

from mcipc.rcon.client import Client


__all__ = ['kill']


def kill(self: Client, targets: str):
    """Kills the targets."""

    return self.run('kill', targets)



"""Enumerations."""

from enum import Enum

from mcipc.rcon.item import Item


__all__ = [
    'Ability',
    'Action',
    'Anchor',
    'Attribute',
    'BossbarSlot',
    'CamerashakeType',
    'CloneMode',
    'Color',
    'DatapackMode',
    'DatapackState',
    'DataType',
    'DebugCommand',
    'Difficulty',
    'Direction',
    'Edition',
    'FillMode',
    'GameMode',
    'Hand',
    'Item',
    'MaskMode',
    'ScanMode',
    'SetblockMode',
    'StorageType',
    'Style',
    'TargetSelector',
    'TargetType',
    'TimeType',
    'XPUnit'
]


class Ability(str, Enum):
    """Player ability."""

    WORLDBUILDER = 'worldbuilder'
    MAYFLY = 'mayfly'
    MUTE = 'mute'


class Action(str, Enum):
    """Modifier action."""

    ADD = 'add'
    MULTIPLY = 'multiply'
    MULTIPLY_BASE = 'multiply_base'


class Anchor(str, Enum):
    """Available entity anchors."""

    EYES = 'eyes'
    FEET = 'feet'


class Attribute(str, Enum):
    """Boss bar attributes."""

    COLOR = 'color'
    MAX = 'max'
    NAME = 'name'
    PLAYERS = 'players'
    STYLE = 'style'
    VALUE = 'value'
    VISIBLE = 'visible'


class BossbarSlot(str, Enum):
    """A target slot off a boss bar."""

    VALUE = 'value'
    MAX = 'max'


class CamerashakeType(str, Enum):
    """Camera shake types."""

    POSITIONAL = 'positional'
    ROTATIONAL = 'rotational'


class CloneMode(str, Enum):
    """Clone modes."""

    FORCE = 'force'
    MOVE = 'move'
    NORMAL = 'normal'


class Color(str, Enum):
    """Boss bar colors."""

    BLUE = 'blue'
    GREEN = 'green'
    PINK = 'pink'
    PURPLE = 'purple'
    RED = 'red'
    WHITE = 'white'
    YELLOW = 'yellow'


class DatapackMode(str, Enum):
    """Available modes."""

    AFTER = 'after'
    BEFORE = 'before'
    FIRST = 'first'
    LAST = 'last'


class DatapackState(str, Enum):
    """Available states."""

    AVAILABLE = 'available'
    ENABLED = 'enabled'


class DataType(str, Enum):
    """Data types."""

    BYTE = 'byte'
    DOUBLE = 'double'
    FLOAT = 'float'
    INT = 'int'
    LONG = 'long'
    SHORT = 'short'


class DebugCommand(str, Enum):
    """Available debug commands."""

    START = 'start'
    STOP = 'stop'
    REPORT = 'report'


class Difficulty(str, Enum):
    """Available difficulties."""

    PEACEFUL = 'peaceful'
    EASY = 'easy'
    NORMAL = 'normal'
    HARD = 'hard'


class Direction(str, Enum):
    """Specifies the direction of Agent to destroy."""

    FORWARD = 'forward'
    BACK = 'back'
    LEFT = 'left'
    RIGHT = 'right'
    UP = 'up'
    DOWN = 'down'


class Edition(str, Enum):
    """Represents the several Minecraft editions."""

    BEDROCK = 'Bedrock Edition'
    EDUCATION = 'Education Edition'
    JAVA = 'Java Edition'


class FillMode(str, Enum):
    """Available fill modes."""

    DESTROY = 'destroy'
    HOLLOW = 'hollow'
    KEEP = 'keep'
    OUTLINE = 'outline'
    REPLACE = 'replace'


class GameMode(str, Enum):
    """Available game modes."""

    ADVENTURE = 'adventure'
    CREATIVE = 'creative'
    SPECTATOR = 'spectator'
    SURVIVAL = 'survival'


class Hand(str, Enum):
    """Available hands."""

    MAINHAND = 'mainhand'
    OFFHAND = 'offhand'


class MaskMode(str, Enum):
    """Masking mode."""

    REPLACE = 'replace'
    MASKED = 'masked'
    FILTERED = 'filtered'


class ScanMode(str, Enum):
    """Available scan modes."""

    ALL = 'all'
    MASKED = 'masked'


class SetblockMode(str, Enum):
    """Available modes to set blocks."""

    DESTROY = 'destroy'
    KEEP = 'keep'
    REPLACE = 'replace'


class StorageType(str, Enum):
    """Storage types."""

    RESULT = 'result'
    SUCCESS = 'success'


class Style(str, Enum):
    """Available boss bar styles."""

    NOTCHED_6 = 'notched_6'
    NOTCHED_10 = 'notched_10'
    NOTCHED_12 = 'notched_12'
    NOTCHED_20 = 'notched_20'
    PROGRESS = 'progress'


class TargetSelector(str, Enum):
    """A target selector."""

    AGENT = '@c'
    ALL_AGENTS = '@v'
    ALL_ENTITIES = '@e'
    ALL_PLAYERS = '@a'
    ISSUER = '@s'
    NEAREST_PLAYER = '@p'
    RANDOM_PLAYER = '@r'


class TargetType(str, Enum):
    """A target type."""

    BLOCK = 'block'
    ENTITY = 'entity'
    STORAGE = 'storage'


class TimeType(str, Enum):
    """Available time types."""

    DAYTIME = 'daytime'
    GAMETIME = 'gametime'
    DAY = 'day'


class XPUnit(str, Enum):
    """Available experience units."""

    LEVELS = 'levels'
    POINTS = 'points'



"""Types for client -> Bedrock Edition server communication."""

from typing import Union

from mcipc.rcon.be.enumerations import Biome
from mcipc.rcon.be.enumerations import Enchantment
from mcipc.rcon.be.enumerations import EntityEquipmentSlot
from mcipc.rcon.be.enumerations import FillType
from mcipc.rcon.be.enumerations import Location
from mcipc.rcon.be.enumerations import MatchMode
from mcipc.rcon.be.enumerations import Mirror
from mcipc.rcon.be.enumerations import MusicRepeatMode
from mcipc.rcon.be.enumerations import Operator
from mcipc.rcon.be.enumerations import Order
from mcipc.rcon.be.enumerations import ReplaceMode
from mcipc.rcon.be.enumerations import RideRules
from mcipc.rcon.be.enumerations import Rotation
from mcipc.rcon.be.enumerations import SaveCommand
from mcipc.rcon.be.enumerations import Structure
from mcipc.rcon.be.enumerations import StructureAnimationMode
from mcipc.rcon.be.enumerations import StructureSaveMode
from mcipc.rcon.be.enumerations import TeleportRules
from mcipc.rcon.be.enumerations import TimeSpec


__all__ = [
    'Biome',
    'Enchantment',
    'EntityEquipmentSlot',
    'FillType',
    'Location',
    'MatchMode',
    'Mirror',
    'MusicRepeatMode',
    'Operator',
    'Order',
    'RelativeFloat',
    'ReplaceMode',
    'RideRules',
    'Rotation',
    'SaveCommand',
    'Structure',
    'StructureAnimationMode',
    'StructureSaveMode',
    'TeleportRules',
    'TimeSpec'
]


Biome = Union[Biome, str]
Enchantment = Union[Enchantment, str]
EntityEquipmentSlot = Union[EntityEquipmentSlot, str]
FillType = Union[FillType, str]
Location = Union[Location, str]
MatchMode = Union[MatchMode, str]
Mirror = Union[Mirror, str]
MusicRepeatMode = Union[MusicRepeatMode, str]
Operator = Union[Operator, str]
Order = Union[Order, str]
RelativeFloat = Union[float, str]
ReplaceMode = Union[ReplaceMode, str]
RideRules = Union[RideRules, str]
Rotation = Union[Rotation, str]
SaveCommand = Union[SaveCommand, str]
Structure = Union[Structure, str]
StructureAnimationMode = Union[StructureAnimationMode, str]
StructureSaveMode = Union[StructureSaveMode, str]
TeleportRules = Union[TeleportRules, str]
TimeSpec = Union[TimeSpec, str]



"""Implementation of the schedule command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.je.types import ScheduleMode, Time
from mcipc.rcon.proxy import CommandProxy


__all__ = ['ScheduleProxy', 'schedule']


class ScheduleProxy(CommandProxy):
    """Proxy for schedule sub-commands."""

    def function(
            self,
            function: str,
            time: Time,
            mode: Optional[ScheduleMode] = None
    ) -> str:
        """Adds a schedule."""
        return self._run('function', function, time, mode)

    def clear(self, function: str) -> str:
        """Removes a schedule."""
        return self._run('clear', function)


def schedule(self: Client) -> ScheduleProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.schedule.ScheduleProxy`
    """

    return ScheduleProxy(self, 'schedule')



"""Implementation of the worldborder command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import Vec2


__all__ = ['DamageProxy', 'WarningProxy', 'WorldborderProxy', 'worldborder']


class DamageProxy(CommandProxy):
    """Proxy for damage commands."""

    def amount(self, damage_per_block: float) -> str:
        """Sets the amount of damage per block."""
        return self._run('amount', damage_per_block)

    def buffer(self, distance: float) -> str:
        """Sets the distance for the damage buffer."""
        return self._run('buffer', distance)


class WarningProxy(CommandProxy):
    """Proxy for warning commands."""

    def distance(self, distance: int) -> str:
        """Sets the warning distance."""
        return self._run('distance', distance)

    def time(self, time: int) -> str:
        """Sets the warning time."""
        return self._run('time', time)


class WorldborderProxy(CommandProxy):
    """Proxy for worldborder commands."""

    def add(self, distance: float, time: Optional[int] = None) -> str:
        """Adds a world border."""
        return self._run('add', distance, time)

    def center(self, pos: Vec2) -> str:
        """Sets the center of the world border."""
        return self._run('center', pos)

    @property
    def damage(self) -> DamageProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.worldborder.DamageProxy`
        """
        return self._proxy(DamageProxy, 'damage')

    def get(self) -> str:
        """Returns information about the world border."""
        return self._run('get')

    def set(self, distance: float, time: Optional[int] = None) -> str:
        """Sets the world border."""
        return self._run('set', distance, time)

    @property
    def warning(self) -> WarningProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.worldborder.WarningProxy`
        """
        return self._proxy(WorldborderProxy, 'warning')


def worldborder(self: Client) -> WorldborderProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.worldborder.WorldborderProxy`
    """

    return WorldborderProxy(self, 'worldborder')



"""Implementation of the ride command."""

from typing import Optional

from mcipc.rcon.be.types import FillType, RideRules, TeleportRules
from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy


__all__ = ['RideProxy', 'ride']


class RideProxy(CommandProxy):
    """Proxy for ride related commands."""

    def start_riding(
            self,
            ride: str,
            teleport_rules: Optional[TeleportRules] = None,
            how_to_fill: Optional[FillType] = None
    ) -> str:
        """Makes <riders> ride on <ride>."""
        return self._run('start_riding', ride, teleport_rules, how_to_fill)

    def stop_riding(self) -> str:
        """Makes <riders> dismount."""
        return self._run('stop_riding')

    def evict_riders(self) -> str:
        """Makes entities that are riding on <rides> dismount."""
        return self._run('evict_riders')

    def summon_rider(
            self,
            entity_type: str,
            spawn_event: Optional[str] = None,
            name_tag: Optional[str] = None
    ) -> str:
        """Summons a rider."""
        return self._run('summon_rider', entity_type, spawn_event, name_tag)

    def summon_ride(
            self,
            entity_type: str,
            ride_rules: Optional[RideRules] = None,
            spawn_event: Optional[str] = None,
            name_tag: Optional[str] = None
    ) -> str:
        """Summons a ride."""
        return self._run('summon_ride', entity_type, ride_rules, spawn_event,
                         name_tag)


def ride(self: Client, riders_or_rides: str) -> RideProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.be.commands.ride.RideProxy`
    """

    return RideProxy(self, 'ride', riders_or_rides)



"""Implementation of the publish command."""

from typing import Optional

from mcipc.rcon.client import Client


__all__ = ['publish']


def publish(self: Client, port: Optional[int] = None) -> str:
    """Opens singleplayer world to the local network."""

    return self.run('publish', port)



"""Implementation of the list command."""

from mcipc.rcon.client import Client
from mcipc.rcon.functions import parsed
from mcipc.rcon.response_types.players import parse


__all__ = ['list']


@parsed(parse)
def list(client: Client, uuids: bool = False) -> str:
    """Returns the players."""

    return client.run('list', 'uuids' if uuids else None)



"""Information about online players."""

from re import Match, fullmatch
from typing import Iterator, NamedTuple, Optional
from uuid import UUID

from mcipc.functions import json_serializable


__all__ = ['Player', 'Players', 'parse']


REGEX_JAVA = r'.+ (\d+) .+ (\d+) .+: (.*)'
REGEX_JAVA_NAME = r'(\S+)(?: \((\S+)\))?'
REGEX_PAPER = r'.+ §c(\d+)§6 .+ §c(\d+)§6 .+\.([\s\S]*)'
REGEX_PAPER_NAME = r'§6(.+)§r: (?:§4)?(\w+)(?:§r)?§f'


@json_serializable
class Player(NamedTuple):
    """Player names with optional UUIDs."""

    name: str
    uuid: Optional[UUID] = None
    state: Optional[str] = None


@json_serializable
class Players(NamedTuple):
    """Online players information."""

    online: int
    max: int
    players: list[Player]

    @property
    def names(self) -> list[str]:
        """Returns a list of the players' names
        for backward compatibility.
        """
        return [player.name for player in self.players]


def player_from_java_name(name: str) -> Player:
    """Returns a player from a Java Edition response name."""

    if (match := fullmatch(REGEX_JAVA_NAME, name)) is None:
        raise ValueError(f'Invalid Java Edition server string: {name}')

    name, uuid = match.groups()
    return Player(name, uuid=None if uuid is None else UUID(uuid))


def players_from_java_names(names: str) -> Iterator[Player]:
    """Yields players from a Java Edition response names."""

    for name in filter(None, map(str.strip, names.split(', '))):
        yield player_from_java_name(name)


def player_from_paper_name(name: str) -> Player:
    """Returns a player from a Paper server response name."""

    if (match := fullmatch(REGEX_PAPER_NAME, name.strip())) is None:
        raise ValueError(f'Invalid Paper server string: {name}')

    state, name = match.groups()
    return Player(name, state=state)


def players_from_paper_names(names: str) -> Iterator[Player]:
    """Yields players from a paper server response names."""

    for name in filter(None, names.split('\n')):
        yield player_from_paper_name(name)


def from_java(match: Match) -> Players:
    """Creates the players information from a Java Edition server match."""

    online, max_, names = match.groups()
    players = players_from_java_names(names)
    return Players(int(online), int(max_), list(players))


def from_paper(match: Match) -> Players:
    """Creates the players information from a Paper server match.
    https://github.com/conqp/mcipc/issues/13#issuecomment-726145034
    """

    online, max_, names = match.groups()
    players = players_from_paper_names(names)
    return Players(int(online), int(max_), list(players))


def parse(text: str) -> Players:
    """Creates the players information from a server response."""

    if (match := fullmatch(REGEX_JAVA, text)) is not None:
        return from_java(match)

    if (match := fullmatch(REGEX_PAPER, text)) is not None:
        return from_paper(match)

    raise ValueError('Unexpected players response:', text)



"""Implementation of the stopsound command."""

from typing import Optional

from mcipc.rcon.client import Client


__all__ = ['stopsound']


def stopsound(self: Client, player: str, sound: Optional[str] = None) -> str:
    """Stops playing a sound."""

    return self.run('stopsound', player, sound)



"""Implementation of the enchant command."""

from typing import Optional, Union

from mcipc.rcon.be.types import Enchantment
from mcipc.rcon.client import Client


__all__ = ['enchant']


def enchant(
        self: Client,
        player: str,
        enchantment: Union[Enchantment, int],
        level: Optional[int] = None
) -> str:
    """Enchants the given player."""

    return self.run('enchant', player, enchantment, level)



"""Common weather proxy implementation."""

from typing import Optional

from mcipc.rcon.proxy import CommandProxy


__all__ = ['WeatherProxy']


class WeatherProxy(CommandProxy):
    """Proxy for weather commands."""

    def clear(self, duration: Optional[int] = None) -> str:
        """Sets the weather to clear."""
        return self._run('clear', duration)

    def rain(self, duration: Optional[int] = None) -> str:
        """Sets the weather to raining."""
        return self._run('rain', duration)

    def thunder(self, duration: Optional[int] = None) -> str:
        """Sets the weather to a thunderstorm."""
        return self._run('thunder', duration)



"""Implementation of the locate command."""

from mcipc.rcon.client import Client
from mcipc.rcon.functions import parsed
from mcipc.rcon.je.types import Structure
from mcipc.rcon.response_types.location import parse


__all__ = ['locate']


@parsed(parse)
def locate(client: Client, structure: Structure) -> str:
    """Locates the respective structure."""

    return client.run('locate', structure)



"""Implementation of the position command."""

from mcipc.rcon.client import Client


__all__ = ['position']


def position(self: Client) -> str:
    """Toggles player coordinates to be displayed on the HUD."""

    return self.run('position')



"""Implementation of the weather command."""

from mcipc.rcon.client import Client
from mcipc.rcon.commands.weather import WeatherProxy


__all__ = ['WeatherProxy', 'weather']


def weather(self: Client) -> WeatherProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.weather.WeatherProxy`
    """

    return WeatherProxy(self, 'weather')



"""Implementation of the team command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.je.types import TeamOption, TeamValue
from mcipc.rcon.proxy import CommandProxy


__all__ = ['TeamProxy', 'team']


class TeamProxy(CommandProxy):
    """Proxy for team commands."""

    # pylint: disable=W0621
    def add(self, team: str, display_name: Optional[dict] = None) -> str:
        """Adds a team."""
        return self._run('add', team, display_name)

    def empty(self, team: str) -> str:
        """Emties a team."""
        return self._run('empty', team)

    def join(self, team: str, members: Optional[str] = None) -> str:
        """Joins a team."""
        return self._run('join', team, members)

    def leave(self, members: Optional[str] = None) -> str:
        """Leaves the team."""
        return self._run('leave', members)

    def list(self, team: Optional[str] = None) -> str:
        """Lists team members."""
        return self._run('list', team)

    def modify(self, team: str, option: TeamOption, value: TeamValue) -> str:
        """Modifies a team."""
        return self._run('modify', team, option, value)

    def remove(self, team: str) -> str:
        """Removes the given team."""
        return self._run('remove', team)


def team(self: Client) -> TeamProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.team.TeamProxy`
    """

    return TeamProxy(self, 'team')



"""Implementation of the fog command."""

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy


__all__ = ['FogProxy', 'fog']


class FogProxy(CommandProxy):
    """Proxy for fog-related commands."""

    def push(self, fog_id: str, user_provided_id: str) -> str:
        """Adds the respective fog."""
        return self._run('push', fog_id, user_provided_id)

    def pop(self, user_provided_id: str) -> str:
        """Pops the respective fog."""
        return self._run('pop', user_provided_id)

    def remove(self, user_provided_id: str) -> str:
        """Remove the respective fog."""
        return self._run('remove', user_provided_id)


def fog(self: Client, victim: str) -> FogProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.be.commands.fog.Proxy`
    """

    return FogProxy(self, 'fog', victim)



"""Implementation of the save-* commands."""

from mcipc.rcon.client import Client


__all__ = ['save_all', 'save_off', 'save_on']


def save_all(self: Client, flush: bool = False) -> str:
    """Saves the server to the data storage device."""

    return self.run('save-all', 'flush' if flush else None)


def save_off(self: Client) -> str:
    """Disables the server writing to the world files."""

    return self.run('save-off')


def save_on(self: Client) -> str:
    """Enables the server writing to the world files."""

    return self.run('save-on')



"""Implementation of the setblock command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.types import SetblockMode, Vec3


__all__ = ['setblock']


def setblock(
        self: Client,
        pos: Vec3,
        block: str,
        mode: Optional[SetblockMode] = None
) -> str:
    """Sets a block."""

    return self.run('setblock', pos, block, mode)



"""Implementation of the summon command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.types import Vec3


__all__ = ['summon']


def summon(
        self: Client,
        entity: str,
        pos: Optional[Vec3] = None,
        nbt: Optional[dict] = None
) -> str:
    """Summons an entity."""

    return self.run('summon', entity, pos, nbt)



"""Implements the testforblocks command."""

from typing import Optional

from mcipc.rcon.be.types import MatchMode
from mcipc.rcon.client import Client
from mcipc.rcon.types import Vec3


__all__ = ['testforblocks']


def testforblocks(
        self: Client,
        begin: Vec3,
        end: Vec3,
        destination: Vec3,
        match_mode: Optional[MatchMode] = None
) -> str:
    """Tests whether the blocks in two regions match."""

    return self.run('testforblocks', begin, end, destination, match_mode)



"""Parses a dict from a response text."""

from collections import defaultdict


__all__ = ['parse']


def parse(text: str) -> dict:
    """Creates the help object from a server response text."""

    help_ = defaultdict(list)

    for use_case in filter(None, text.split('/')):
        command, *arguments = use_case.split()
        help_[command].append(arguments)

    return help_



"""Implementation of the transferserver command."""

from mcipc.rcon.client import Client


__all__ = ['transferserver']


def transferserver(self: Client, server: str, port: int) -> str:
    """Transfer player to another server."""

    return self.run('transferserver', server, port)



"""Implementation of the setidletimeout command."""

from mcipc.rcon.client import Client


__all__ = ['setidletimeout']


def setidletimeout(self: Client, minutes: int) -> str:
    """Sets the time before idle players are kicked from the server."""

    return self.run('setidletimeout', minutes)



"""Implementation of the setworldspawn command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.types import Vec3


__all__ = ['setworldspawn']


def setworldspawn(
        self: Client,
        pos: Optional[Vec3] = None,
        angle: Optional[float] = None
) -> str:
    """Sets the world spawn."""

    return self.run('setworldspawn', pos, angle)



"""Implementation of the debug command."""

from mcipc.rcon.client import Client
from mcipc.rcon.types import DebugCommand


__all__ = ['debug']


def debug(self: Client, command: DebugCommand) -> str:
    """Invokes the debug command."""

    return self.run('debug', command)



"""Implementation of the enchant command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.je.types import Enchantment


__all__ = ['enchant']


def enchant(
        self: Client,
        target: str,
        enchantment: Enchantment,
        level: Optional[int] = None
) -> str:
    """Enchants the target."""

    return self.run('enchant', target, enchantment, level)



"""Implementation of the help command."""

from typing import Optional, Union

from mcipc.rcon.client import Client


__all__ = ['help']


# pylint: disable=W0622
def help(
        client: Client,
        command_or_page: Optional[Union[str, int]] = None
) -> str:
    """Returns help about commands."""

    return client.run('help', command_or_page)



"""Implementation of the playanimation command."""

from typing import Optional

from mcipc.rcon.client import Client


__init__ = [' playanimation']


# pylint: disable=R0913
def playanimation(
        self: Client,
        entity: str,
        animation: str,
        next_state: Optional[str] = None,
        blend_out_time: Optional[float] = None,
        stop_expression: Optional[str] = None,
        controller: Optional[str] = None
) -> str:
    """Makes one or more entities play a one-off animation."""

    return self.run('playanimation', entity, animation, next_state,
                    blend_out_time, stop_expression, controller)



"""Implementation of the spreadplayers command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.types import Vec2


__all__ = ['spreadplayers']


# pylint: disable=R0913
def spreadplayers(
        self: Client,
        center: Vec2,
        spread_distance: float,
        max_range: float,
        respec_teams: bool,
        targets: str, *,
        max_height: Optional[int] = None
) -> str:
    """Spreads players to resulting position under the maximum height."""

    command = ['spreadplayers', center, spread_distance, max_range]

    if max_height is not None:
        command += ['under', max_height]

    return self.run(*command, respec_teams, targets)



"""Implementation of the geteduclientinfo command."""

from mcipc.rcon.client import Client


__all__ = ['geteduclientinfo']


def geteduclientinfo(self: Client) -> str:
    """Returns information about the game's client."""

    return self.run('geteduclientinfo')



"""Implementation of the weather command."""

from mcipc.rcon.client import Client
from mcipc.rcon.commands.weather import WeatherProxy as _WeatherProxy


__all__ = ['WeatherProxy', 'weather']


class WeatherProxy(_WeatherProxy):
    """Proxy for weather commands."""

    def query(self) -> str:
        """Queries the current weather."""
        return self._run('query')


def weather(self: Client) -> WeatherProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.be.commands.weather.WeatherProxy`
    """

    return WeatherProxy(self, 'weather')



"""Implementation of the particle command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.je.types import ParticleMode
from mcipc.rcon.types import Vec3


__all__ = ['particle']


# pylint: disable=R0913
def particle(
        self: Client,
        name: str,
        speed: float,
        count: int,
        parameters: Optional[str] = None,
        pos: Optional[Vec3] = None,
        delta: Optional[Vec3] = None,
        mode: Optional[ParticleMode] = None,
        viewers: Optional[str] = None
) -> str:
    """Creates the respective particles."""

    args = ['particle', name]

    if parameters is not None:
        args.append(parameters)

    if pos is not None:
        args.append(pos)

        if delta is not None:
            args.append(delta)

    return self.run(*args, speed, count, mode, viewers)



"""Basic statistics protocol."""

from __future__ import annotations
from typing import IO, NamedTuple

from mcipc.functions import json_serializable
from mcipc.query.proto.common import MAGIC
from mcipc.query.proto.common import NULL
from mcipc.query.proto.common import decode_all
from mcipc.query.proto.common import ip_or_hostname
from mcipc.query.proto.common import random_session_id
from mcipc.query.proto.common import BigEndianSignedInt32
from mcipc.query.proto.common import IPAddressOrHostname
from mcipc.query.proto.common import Type


__all__ = ['Request', 'BasicStats']


class Request(NamedTuple):
    """Basic statistics request packet."""

    magic: bytes = MAGIC
    type: Type = Type.STAT
    session_id: BigEndianSignedInt32 = BigEndianSignedInt32()
    challenge_token: BigEndianSignedInt32 = BigEndianSignedInt32()

    def __bytes__(self):
        """Returns the packet as bytes."""
        payload = self.magic
        payload += bytes(self.type)
        payload += bytes(self.session_id)
        payload += bytes(self.challenge_token)
        return payload

    @classmethod
    def create(cls, challenge_token: BigEndianSignedInt32) -> Request:
        """Creates a new request with the given challenge token."""
        return cls(session_id=random_session_id(),
                   challenge_token=challenge_token)


@json_serializable
class BasicStats(NamedTuple):
    """Basic statistics response packet."""

    type: Type
    session_id: BigEndianSignedInt32
    motd: str
    game_type: str
    map: str
    num_players: int
    max_players: int
    host_port: int
    host_ip: IPAddressOrHostname

    @classmethod
    def read(cls, file: IO) -> BasicStats:
        """Reads the basic stats from a file-like object."""
        type_ = Type.read(file)
        session_id = BigEndianSignedInt32.read(file)
        body = b''

        while True:
            body += file.read(1)

            if len(body.split(NULL)) == 7:
                break

        *blocks, port_ip, _ = body.split(NULL)
        motd, game_type, map_, num_players, max_players = decode_all(blocks)
        num_players = int(num_players)
        max_players = int(max_players)
        host_port = int.from_bytes(port_ip[0:2], 'little')
        host_ip = ip_or_hostname(port_ip[2:].decode())
        return cls(
            type_, session_id, motd, game_type, map_, num_players, max_players,
            host_port, host_ip)



"""Minecraft-specific client."""

from rcon import source

from mcipc.rcon.functions import str_until_none


__all__ = ['Client']


class Client(source.Client, frag_detect='seed'):
    """An RCON client for Minecraft."""

    def run(self, command: str, *arguments: str) -> str:
        """Runs the command with additional checks."""
        return super().run(*str_until_none(command, *arguments))



"""Common functions."""

from typing import Any, Union


__all__ = ['json_serializable']


JSON = Union[dict, list, float, int, str, bool, None]


def jsonify(value: Any) -> JSON:
    """Converts a value into a JSON-compliant value."""

    if value is None:
        return None

    if isinstance(value, dict):
        return {key: jsonify(value) for key, value in value.items()}

    if isinstance(value, list):
        return [jsonify(item) for item in value]

    if isinstance(value, (float, int, str, bool)):
        return value

    try:
        value = dict(value)
    except (TypeError, ValueError):
        return str(value)

    return jsonify(value)


def get_json_item(instance: type, index_or_key: Union[int, str]) -> Any:
    """Returns the respective item by index or field name.
    In the case of a field name, the item will be converted
    into a JSON-ish value.
    """

    if isinstance(index_or_key, str):
        try:
            value = getattr(instance, index_or_key)
        except AttributeError:
            raise KeyError(index_or_key) from None

        return jsonify(value)

    return super().__getitem__(index_or_key)


def json_serializable(cls: type) -> type:
    """Extends a NamedTuple with methods to return a JSON-ish
    dict when passed to the dict() constructor.
    """

    if not hasattr(cls, 'keys'):
        if not (issubclass(cls, tuple) and hasattr(cls, '_fields')):
            raise TypeError('Type does not implement method keys(self):', cls)

        cls.keys = lambda self: self._fields

    cls.__getitem__ = get_json_item
    return cls



"""Implementation of the spectate command."""

from typing import Optional

from mcipc.rcon.client import Client


__all__ = ['spectate']


def spectate(
        self: Client,
        target: Optional[str] = None,
        player: Optional[str] = None
) -> str:
    """Start or stop spectating."""

    return self.run('spectate', target, player)



"""Implementation of the clone command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.enumerations import MaskMode
from mcipc.rcon.types import CloneMode, MaskMode as MaskModeHint, Vec3


__all__ = ['clone']


def clone(
        self: Client,
        begin: Vec3,
        end: Vec3,
        destination: Vec3,
        *,
        mask_mode: Optional[MaskModeHint] = None,
        clone_mode: Optional[CloneMode] = None,
        tile_name: Optional[str] = None,
        tile_data: Optional[int] = None
) -> str:
    """Clones blocks from one region to another."""

    args = ['clone', begin, end, destination]

    if mask_mode is not None:
        args.append(mask_mode := MaskMode(mask_mode))

        if filtered := mask_mode == MaskMode.FILTERED:
            if clone_mode is None:
                raise ValueError('Missing clone mode.')

            if tile_name is None:
                raise ValueError('Missing tile name.')

            if tile_data is None:
                raise ValueError('Missing tile data.')

        args.append(clone_mode)

        if filtered:
            args += [tile_name, tile_data]

    return self.run(*args)



"""Implementation of the locate command."""

from mcipc.rcon.be.types import Structure
from mcipc.rcon.client import Client


__all__ = ['locate']


def locate(client: Client, structure: Structure) -> str:
    """Locates the respective structure."""

    return client.run('locate', structure)



"""Implementation of the teammsg and tm commands."""

from mcipc.rcon.client import Client


__all__ = ['teammsg', 'tm']


def teammsg(self: Client, message: str) -> str:
    """Sends a team message."""

    return self.run('teammsg', message)


def tm(self: Client, message: str) -> str:  # pylint: disable=C0103
    """Sends a team message."""

    return self.run('tm', message)



"""Implementation of the setworldspawn command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.types import Vec3


__all__ = ['setworldspawn']


def setworldspawn(self: Client, spawn_point: Optional[Vec3] = None) -> str:
    """Sets the world spawn."""

    return self.run('setworldspawn', spawn_point)



"""Implementation of the trigger command."""

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy


__all__ = ['TriggerProxy', 'trigger']


class TriggerProxy(CommandProxy):
    """Proxy for trigger commands."""

    def __call__(self, objective: str) -> str:
        """Adds 1 to the current value of <objective>."""
        return self._run('objective')

    def add(self, objective: str, value: int) -> str:
        """Adds <value> to the current value of <objective>."""
        return self._run('add', objective, value)

    def set(self, objective: str, value: int) -> str:
        """Sets the value of <objective> to <value>."""
        return self._run('set', objective, value)


def trigger(self: Client) -> TriggerProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.trigger.TriggerProxy`
    """

    return TriggerProxy(self, 'trigger')



"""Implementation of the code command."""

from mcipc.rcon.client import Client


__all__ = ['code']


def code(self: Client) -> str:
    """Used to access code connection."""

    return self.run('code')



"""Implementation of the seed command."""

from mcipc.rcon.client import Client
from mcipc.rcon.functions import parsed
from mcipc.rcon.response_types.seed import parse


__all__ = ['seed']


@parsed(parse)
def seed(self: Client) -> str:
    """Returns the server seed."""

    return self.run('seed')



"""Banning, pardoning and kicking of players or IP addresses."""

from mcipc.rcon.client import Client
from mcipc.rcon.types import IPAddressOrHostname


__all__ = ['ban', 'ban_ip', 'banlist', 'pardon']


def ban(self: Client, player: str, *reasons: str) -> str:
    """Adds a player to then ban list."""

    return self.run('ban', player, *reasons)


def ban_ip(self: Client, target: IPAddressOrHostname, *reasons: str) -> str:
    """Adds an IP address to the ban list."""

    return self.run('ban-ip', target, *reasons)


def banlist(self: Client, *ips_or_players: IPAddressOrHostname) -> str:
    """Displays the server's ban list."""

    return self.run('banlist', *ips_or_players)


def pardon(self: Client, target: str) -> str:
    """Removes entries from the ban list."""

    return self.run('pardon', target)



"""Enumerations specific to the Java Edition."""

from enum import Enum


__all__ = [
    'Biome',
    'CollisionRule',
    'Color',
    'Enchantment',
    'ParticleMode',
    'RenderType',
    'ScheduleMode',
    'SoundSource',
    'Structure',
    'TeamOption',
    'TimeSpec',
    'TimeUnit',
    'Visibility'
]


class Biome(str, Enum):
    """Available biomes in the Java Edition."""

    BADLANDS = 'badlands'
    BADLANDS_PLATEAU = 'badlands_plateau'
    BAMBOO_JUNGLE = 'bamboo_jungle'
    BAMBOO_JUNGLE_HILLS = 'bamboo_jungle_hills'
    BASALT_DELTAS = 'basalt_deltas'
    BEACH = 'beach'
    BIRCH_FOREST = 'birch_forest'
    BIRCH_FOREST_HILLS = 'birch_forest_hills'
    COLD_OCEAN = 'cold_ocean'
    CRIMSON_FOREST = 'crimson_forest'
    DARK_FOREST = 'dark_forest'
    DARK_FOREST_HILLS = 'dark_forest_hills'
    DEEP_COLD_OCEAN = 'deep_cold_ocean'
    DEEP_FROZEN_OCEAN = 'deep_frozen_ocean'
    DEEP_LUKEWARM_OCEAN = 'deep_lukewarm_ocean'
    DEEP_OCEAN = 'deep_ocean'
    DEEP_WARM_OCEAN = 'deep_warm_ocean'
    DESERT = 'desert'
    DESERT_HILLS = 'desert_hills'
    DESERT_LAKES = 'desert_lakes'
    DRIPSTONE_CAVES = 'dripstone_caves'
    END_BARRENS = 'end_barrens'
    END_HIGHLANDS = 'end_highlands'
    END_MIDLANDS = 'end_midlands'
    ERODED_BADLANDS = 'eroded_badlands'
    FLOWER_FOREST = 'flower_forest'
    FOREST = 'forest'
    FROZEN_OCEAN = 'frozen_ocean'
    FROZEN_RIVER = 'frozen_river'
    GIANT_SPRUCE_TAIGA = 'giant_spruce_taiga'
    GIANT_SPRUCE_TAIGA_HILLS = 'giant_spruce_taiga_hills'
    GIANT_TREE_TAIGA = 'giant_tree_taiga'
    GIANT_TREE_TAIGA_HILLS = 'giant_tree_taiga_hills'
    GRAVELLY_MOUNTAINS = 'gravelly_mountains'
    ICE_SPIKES = 'ice_spikes'
    JUNGLE = 'jungle'
    JUNGLE_EDGE = 'jungle_edge'
    JUNGLE_HILLS = 'jungle_hills'
    LUKEWARM_OCEAN = 'lukewarm_ocean'
    MODIFIED_BADLANDS_PLATEAU = 'modified_badlands_plateau'
    MODIFIED_GRAVELLY_MOUNTAINS = 'modified_gravelly_mountains'
    MODIFIED_JUNGLE = 'modified_jungle'
    MODIFIED_JUNGLE_EDGE = 'modified_jungle_edge'
    MODIFIED_WOODEN_BADLANDS_PLATEAU = 'modified_wooded_badlands_plateau'
    MOUNTAIN_EDGE = 'mountain_edge'
    MOUNTAINS = 'mountains'
    MUSHROOM_FIELD_SHORE = 'mushroom_field_shore'
    MUSHROOM_FIELDS = 'mushroom_fields'
    NETHER_WASTES = 'nether_wastes'
    OCEAN = 'ocean'
    PLAINS = 'plains'
    RIVER = 'river'
    SAVANNA = 'savanna'
    SAVANNA_PLATEAU = 'savanna_plateau'
    SHATTERED_SAVANNA = 'shattered_savanna'
    SHATTERED_SAVANNA_PLATEAU = 'shattered_savanna_plateau'
    SMALL_END_ISLANDS = 'small_end_islands'
    SNOWY_BEACH = 'snowy_beach'
    SNOWY_MOUNTAINS = 'snowy_mountains'
    SNOWY_TAIGA = 'snowy_taiga'
    SNOWY_TAIGA_HILLS = 'snowy_taiga_hills'
    SNOWY_TAIGA_MOUNTAINS = 'snowy_taiga_mountains'
    SNOWY_TUNDRA = 'snowy_tundra'
    SOUL_SAND_VALLEY = 'soul_sand_valley'
    STONE_SHORE = 'stone_shore'
    SUNFLOWER_PLAINS = 'sunflower_plains'
    SWAMP = 'swamp'
    SWAMP_HILLS = 'swamp_hills'
    TAIGA = 'taiga'
    TAIGA_HILLS = 'taiga_hills'
    TAIGA_MOUNTAINS = 'taiga_mountains'
    TALL_BIRCH_FOREST = 'tall_birch_forest'
    TALL_BIRCH_HILLS = 'tall_birch_hills'
    THE_END = 'the_end'
    THE_VOID = 'the_void'
    WARM_OCEAN = 'warm_ocean'
    WARPED_FOREST = 'warped_forest'
    WOODED_BADLANDS_PLATEAU = 'wooded_badlands_plateau'
    WOODED_HILLS = 'wooded_hills'
    WOODED_MOUNTAINS = 'wooded_mountains'


class CollisionRule(str, Enum):
    """Available collision rules."""

    ALWAYS = 'always'
    NEVER = 'never'
    PUSH_OTHER_TEAMS = 'pushOtherTeams'
    PUSH_OWN_TEAM = 'pushOwnTeam'


class Color(str, Enum):
    """Available colors."""

    AQUA = 'aqua'
    BLACK = 'black'
    BLUE = 'blue'
    DARK_AQUA = 'dark_aqua'
    DARK_BLUE = 'dark_blue'
    DARK_GRAY = 'dark_gray'
    DARK_GREEN = 'dark_green'
    DARK_PURPLE = 'dark_purple'
    DARK_RED = 'dark_red'
    GOLD = 'gold'
    GRAY = 'gray'
    GREEN = 'green'
    LIGHT_PURPLE = 'light_purple'
    RED = 'red'
    RESET = 'reset'
    YELLOW = 'yellow'
    WHITE = 'white'


class Enchantment(str, Enum):
    """Available enchantments in the Java Edition."""

    AQUA_AFFINITY = 'aqua_affinity'
    BANE_OF_ARTHROPODS = 'bane_of_arthropods'
    BLAST_PROTECTION = 'blast_protection'
    CHANNELING = 'channeling'
    CLEAVING = 'cleaving'
    CURSE_OF_BINDING = 'binding_curse'
    CURSE_OF_VANISHING = 'vanishing_curse'
    DEPTH_STRIDER = 'depth_strider'
    EFFICIENCY = 'efficiency'
    FEATHER_FALLING = 'feather_falling'
    FIRE_ASPECT = 'fire_aspect'
    FIRE_PROTECTION = 'fire_protection'
    FLAME = 'flame'
    FORTUNE = 'fortune'
    FROST_WALKER = 'frost_walker'
    IMPALING = 'impaling'
    INFINITY = 'infinity'
    KNOCKBACK = 'knockback'
    LOOTING = 'looting'
    LOYALTY = 'loyalty'
    LUCK_OF_THE_SEA = 'luck_of_the_sea'
    LURE = 'lure'
    MENDING = 'mending'
    MULTISHOT = 'multishot'
    PIERCING = 'piercing'
    POWER = 'power'
    PROJECTILE_PROTECTION = 'projectile_protection'
    PROTECTION = 'protection'
    PUNCH = 'punch'
    QUICK_CHARGE = 'quick_charge'
    RESPIRATION = 'respiration'
    RIPTIDE = 'riptide'
    SHARPNESS = 'sharpness'
    SILK_TOUCH = 'silk_touch'
    SMITE = 'smite'
    SOUL_SPEED = 'soul_speed'
    SWEEPING = 'sweeping'
    THORNS = 'thorns'
    UNBREAKING = 'unbreaking'


class ParticleMode(str, Enum):
    """Available particle modes."""

    FORCE = 'force'
    NORMAL = 'normal'


class RenderType(str, Enum):
    """Available render types."""

    HEARTS = 'hearts'
    INTEGER = 'integer'


class ScheduleMode(str, Enum):
    """Available schedule modes."""

    APPEND = 'append'
    REPLACE = 'replace'


class SoundSource(str, Enum):
    """Available sound sources."""

    MASTER = 'master'
    MUSIC = 'music'
    RECORD = 'record'
    WEATHER = 'weather'
    BLOCK = 'block'
    HOSTILE = 'hostile'
    NEUTRAL = 'neutral'
    PLAYER = 'player'
    AMBIENT = 'ambient'
    VOICE = 'voice'


class Structure(str, Enum):
    """Available structures in the Java Edition."""

    BASTION_REMNANT = 'bastion_remnant'
    BURIED_TREASURE = 'buried_treasure'
    END_CITY = 'endcity'
    FORTRESS = 'fortress'
    WOODLAND_MANSION = 'mansion'
    MINESHAFT = 'mineshaft'
    MONUMENT = 'monument'
    NETHER_FOSSIL = 'nether_fossil'
    OCEAN_RUINS = 'ocean_ruin'
    PILLAGER_OUTPOST = 'pillager_outpost'
    RUINED_PORTAL = 'ruined_portal'
    SHIPWRECK = 'shipwreck'
    STRONGHOLD = 'stronghold'
    DESERT_PYRAMID = 'desert_pyramid'
    IGLOO = 'igloo'
    JUNGLE_PYRAMID = 'jungle_pyramid'
    SWAMP_HUT = 'swamp_hut'
    VILLAGE = 'village'


class TeamOption(str, Enum):
    """Available team options."""

    COLLISION_RULE = 'collisionRule'
    COLOR = 'color'
    DEATH_MESSAGE_VISIBILITY = 'deathMessageVisibility'
    DISPLAY_NAME = 'displayName'
    FRIENDLY_FIRE = 'friendlyFire'
    NAMETAG_VISIBILITY = 'nametagVisibility'
    PREFIX = 'prefix'
    SEE_FRIENDLY_INVISIBLES = 'seeFriendlyInvisibles'
    SUFFIX = 'suffix'


class TimeSpec(str, Enum):
    """Available time specifications."""

    DAY = 'day'
    NIGHT = 'night'
    NOON = 'noon'
    MIDNIGHT = 'midnight'


class TimeUnit(str, Enum):
    """Available time units."""

    DAY = 'd'
    SECOND = 's'
    TICK = 't'


class Visibility(str, Enum):
    """Available visibility types."""

    NEVER = 'never'
    HIDE_FOR_OTHER_TEAMS = 'hideForOtherTeams'
    HIDE_FOR_OWN_TEAM = 'hideForOwnTeam'
    ALWAYS = 'always'



"""Implementation of the teleport command."""

from typing import Optional, Union

from mcipc.rcon.be.types import RelativeFloat
from mcipc.rcon.client import Client
from mcipc.rcon.types import Vec3


__all__ = ['teleport']


def extend_command(
        look_at_position: Vec3,
        look_at_entity: str,
        y_rot: RelativeFloat,
        x_rot: RelativeFloat,
        check_for_blocks: bool
):
    """Extends the command with common sub commands."""

    if look_at_position is not None:
        yield'facing'
        yield look_at_position
    elif look_at_entity is not None:
        yield 'facing'
        yield look_at_entity
    else:
        if y_rot is not None:
            yield y_rot

        if x_rot is not None:
            yield x_rot

    yield check_for_blocks


def teleport(
        self: Client,
        *,
        destination: Optional[Union[Vec3, str]] = None,
        victim: Optional[str] = None,
        check_for_blocks: Optional[bool] = None,
        y_rot: Optional[RelativeFloat] = None,
        x_rot: Optional[RelativeFloat] = None,
        look_at_position: Optional[Vec3] = None,
        look_at_entity: Optional[str] = None
) -> str:
    """Teleports the player."""

    command = ['teleport']

    if victim is not None:
        if destination is None:
            raise TypeError('Must spectify destination on victim.')

        command += [victim, destination, *extend_command(
            look_at_position, look_at_entity, y_rot, x_rot, check_for_blocks)]

    elif destination is not None:
        command += [destination, *extend_command(
            look_at_position, look_at_entity, y_rot, x_rot, check_for_blocks)]
    else:
        raise TypeError('Must specify either destination or victim.')

    return self.run(*command)



"""Implementation of the item command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import Vec3


__all__ = ['CopyProxy', 'ItemProxy', 'ItemSubProxy', 'item']


class CopyProxy(CommandProxy):
    """Proxy for item copy commands."""

    def block(
            self,
            pos: Vec3,
            slot: str,
            modifier: Optional[str] = None
    ) -> str:
        """Copies a block."""
        return self._run('block', pos, slot, modifier)

    def entity(
            self,
            targets: str,
            slot: str,
            modifier: Optional[str] = None
    ) -> str:
        """Copies an entity."""
        return self._run('entity', targets, slot, modifier)


class ItemSubProxy(CommandProxy):
    """Sub-proxy for block and entity related commands."""

    @property
    def copy(self) -> CopyProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.item.CopyProxy`
        """
        return self._proxy(CopyProxy, 'copy')

    def modify(self, modifier: str) -> str:
        """Modifies the item."""
        return self._run('modify', modifier)

    # pylint: disable=W0621
    def replace(self, item: str, count: Optional[int] = None) -> str:
        """Replaces an item."""
        return self._run('replace', item, count)


class ItemProxy(CommandProxy):
    """Proxy for item-related commands."""

    def block(self, pos: Vec3, slot: str) -> ItemSubProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.item.ItemSubProxy`
        """
        return self._proxy(ItemSubProxy, 'block', pos, slot)

    def entity(self, targets: str, slot: str) -> ItemSubProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.item.ItemSubProxy`
        """
        return self._proxy(ItemSubProxy, 'entity', targets, slot)


def item(self: Client) -> ItemProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.item.ItemProxy`
    """

    return ItemProxy(self, 'item')



"""Implementation of the gamemode command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.types import GameMode


__all__ = ['gamemode']


def gamemode(
        self: Client,
        mode: GameMode,
        target: Optional[str] = None
) -> str:
    """Sets the game mode."""

    return self.run('gamemode', mode, target)



"""Implementation of loot command."""

from typing import Optional, Union

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import Hand, Vec3


__all__ = ['LootProxy', 'ReplaceProxy', 'SourceProxy', 'loot']


class SourceProxy(CommandProxy):
    """Proxy for source commands."""

    def fish(
            self,
            loot_table: str,
            pos: Vec3,
            tool_or_hand: Optional[Union[Hand, str]] = None
    ) -> str:
        """Runs the command with the fish source."""
        return self._run('fish', loot_table, pos, tool_or_hand)

    def loot(self, loot_table: str) -> str:
        """Runs the command with a loot source."""
        return self._run('loot', loot_table)

    def kill(self, target: str) -> str:
        """Runs the command with a killed target as source."""
        return self._run('kill', target)

    def mine(
            self,
            pos: Vec3,
            tool_or_hand: Optional[Union[Hand, str]] = None
    ) -> str:
        """Runs the command with a mined resource as source."""
        return self._run('mine', pos, tool_or_hand)


class ReplaceProxy(CommandProxy):
    """Proxy for replace commands."""

    def entity(
            self,
            entities: str,
            slot: str,
            count: Optional[int] = None
    ) -> SourceProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.loot.SourceProxy`
        """
        return self._proxy(SourceProxy, 'entity', entities, slot, count)

    def block(
            self,
            target_pos: Vec3,
            slot: str,
            count: Optional[int] = None
    ) -> SourceProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.loot.SourceProxy`
        """
        return self._proxy(SourceProxy, 'block', target_pos, slot, count)


class LootProxy(CommandProxy):
    """Proxy for target-related commands."""

    def spawn(self, target_pos: Vec3) -> SourceProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.loot.SourceProxy`
        """
        return self._proxy(SourceProxy, 'spawn', target_pos)

    @property
    def replace(self) -> ReplaceProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.loot.ReplaceProxy`
        """
        return self._proxy(ReplaceProxy, 'replace')

    def give(self, players: str) -> SourceProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.loot.SourceProxy`
        """
        return self._proxy(SourceProxy, 'give', players)

    def insert(self, target_pos: Vec3) -> SourceProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.je.commands.loot.SourceProxy`
        """
        return self._proxy(SourceProxy, 'insert', target_pos)


def loot(self: Client) -> LootProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.loot.LootProxy`
    """

    return LootProxy(self, 'loot')



"""Implementation of the immutableworld command."""

from mcipc.rcon.client import Client


__all__ = ['immutableworld']


def immutableworld(self: Client, value: bool) -> str:
    """Toggles the world to be able altered or cannot be altered."""

    return self.run('immutableworld', value)



"""Implementation of the stopsound command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.je.types import SoundSource


__all__ = ['stopsound']


def stopsound(
        self: Client,
        targets: str,
        source: Optional[SoundSource] = None,
        sound: Optional[str] = None
) -> str:
    """Stops a sound."""

    return self.run('stopsound', targets, source, sound)



"""Implementation of th gamerule command."""

from typing import Optional, Union

from mcipc.rcon.client import Client


__all__ = ['gamerule']


def gamerule(
        self: Client,
        rule: str,
        value: Optional[Union[bool, int]] = None
) -> str:
    """Sets or queries a game rule value."""

    return self.run('gamerule', rule, value)



"""Implementation of the worldbuilder and wb commands."""

from mcipc.rcon.client import Client


__all__ = ['wb', 'worldbuilder']


def wb(self: Client) -> str:
    """Toggles the world builder feature."""

    return self.run('wb')


def worldbuilder(self: Client) -> str:
    """Toggles the world builder feature."""

    return self.run('worldbuilder')



"""Locations."""

from re import fullmatch
from typing import NamedTuple, Optional

from mcipc.functions import json_serializable
from mcipc.rcon.exceptions import LocationNotFound


__all__ = ['Location', 'parse']


REGEX = (
    r'The nearest (.+) is at \[(-?\d+), (~|-?\d+), '
    r'(-?\d+)\] \((\d+) block[s]? away\)'
)


@json_serializable
class Location(NamedTuple):
    """A 3D location."""

    name: str
    x: int
    y: Optional[int]
    z: int
    distance: int


def parse(text: str) -> Location:
    """Creates a location from a server response."""

    if (match := fullmatch(REGEX, text)) is None:
        raise LocationNotFound(text)

    name, x, y, z, distance = match.groups()    # pylint: disable=C0103
    return Location(name, int(x), None if y == '~' else int(y), int(z),
                    int(distance))



"""Implementation of the tag command."""

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy


__all__ = ['TagProxy', 'tag']


class TagProxy(CommandProxy):
    """Proxy for tag commands."""

    def add(self, name: str) -> str:
        """Adds a tag."""
        return self._run('add', name)

    def list(self) -> str:
        """Lists tags."""
        return self._run('list')

    def remove(self, name: str) -> str:
        """Removes a tag."""
        return self._run('remove', name)


def tag(self: Client, targets: str) -> TagProxy:
    """Delegates to a command proxy."""

    return TagProxy(self, 'tag', targets)



"""Implementation of the scoreboard command."""

from typing import Optional, Union

from mcipc.rcon.be.types import Location, Operator, Order
from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy


__all__ = ['ObjectivesProxy', 'PlayersProxy', 'ScoreboardProxy', 'scoreboard']


class ObjectivesProxy(CommandProxy):
    """Proxy for objective commands."""

    def add(self, objective: str, display_name: Optional[str] = None) -> str:
        """Adds an objective."""
        return self._run('add', objective, 'dummy', display_name)

    def list(self) -> str:
        """Lists objectives."""
        return self._run('list')

    def remove(self, objective: str) -> str:
        """Removes an objective."""
        return self._run('remove', objective)

    def setdisplay(
            self,
            location: Location,
            objective: str,
            order: Optional[Order] = None
    ) -> str:
        """Sets a display."""
        command = ['setdisplay', location := Location(location), objective]

        if location in {Location.LIST, Location.SIDEBAR}:
            command.append(order)

        return self._run(*command)


class PlayersProxy(CommandProxy):
    """Proxy for player commands."""

    def list(self, playername: Optional[str] = None) -> str:
        """Lists player scores."""
        return self._run('list', playername)

    def reset(self, player: str, objective: Optional[str] = None) -> str:
        """Resets a player's score."""
        return self._run('reset', player, objective)

    # pylint: disable=W0622
    def test(
            self,
            player: str,
            objective: str,
            min: Union[int, str],
            max: Optional[Union[int, str]] = None
    ) -> str:
        """Tests whether a player's score is within the given boundaries."""
        return self._run('test', player, objective, min, max)

    def random(
            self,
            player: str,
            objective: str,
            min: int,
            max: Optional[int] = None
    ) -> str:
        """Sets a random score for the player's objective."""
        return self._run('random', player, objective, min, max)

    def set(self, player: str, objective: str, count: int) -> str:
        """Sets a player's score."""
        return self._run('set', player, objective, count)

    def add(self, player: str, objective: str, count: int) -> str:
        """Adds to a player's score."""
        return self._run('add', player, objective, count)

    def remove(self, player: str, objective: str, count: int) -> str:
        """Removes from a player's score."""
        return self._run('remove', player, objective, count)

    # pylint: disable=R0913
    def operation(self, target_name: str, target_objective: str,
                  operation: Operator, selector: str, objective: str) -> str:
        """Executes an operation."""
        return self._run('operation', target_name, target_objective,
                         operation, selector, objective)


class ScoreboardProxy(CommandProxy):
    """Proxy for scoreboard related commands."""

    @property
    def objectives(self) -> ObjectivesProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.be.commands.scoreboard.ObjectivesProxy`
        """
        return self._proxy(ObjectivesProxy, 'objectives')

    @property
    def players(self) -> PlayersProxy:
        """Delegates to a
        :py:class:`mcipc.rcon.be.commands.scoreboard.PlayersProxy`
        """
        return self._proxy(PlayersProxy, 'players')


def scoreboard(self: Client) -> ScoreboardProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.be.commands.scoreboard.ScoreboardProxy`
    """

    return ScoreboardProxy(self, 'scoreboard')



"""Named binary tags."""

from typing import Any, Union

from mcipc.rcon.enumerations import Item, TargetSelector
from mcipc.rcon.functions import stringify


__all__ = ['NBT']


def tags_to_str(tags: dict[str, Any]) -> str:
    """Returns the tags as a string."""

    return ', '.join(f'{key}={stringify(val)}' for key, val in tags.items())


class NBT:
    """Represents a named binary tag."""

    __slots__ = ('target', 'tags')

    def __init__(self, target: Union[Item, TargetSelector, str], **tags):
        """Stores the object name and optional tags."""
        self.target = target
        self.tags = tags

    def __repr__(self):
        """Returns a str representation for eval()."""
        return f'{type(self).__name__}({self.target!r}, {self.tags!r})'

    def __str__(self):
        """Returns a str representation for RCON."""
        return f'{stringify(self.target)}[{tags_to_str(self.tags)}]'



"""Implementation of the time command."""

from typing import Union

from mcipc.rcon.client import Client
from mcipc.rcon.je.types import Time, TimeSpec
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import TimeType


__all__ = ['TimeProxy', 'time']


class TimeProxy(CommandProxy):
    """Proxy for time commands."""

    # pylint: disable=W0621
    def add(self, time: Time) -> str:
        """Adds time."""
        return self._run('add', time)

    def query(self, time_type: TimeType) -> str:
        """Queries the given time type."""
        return self._run('query', time_type)

    def set(self, time: Union[TimeSpec, Time]) -> str:
        """Sets the time to the given amount."""
        return self._run('set', time)


def time(self: Client) -> TimeProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.time.TimeProxy`
    """

    return TimeProxy(self, 'time')



"""Implementation of the difficulty command."""

from mcipc.rcon.client import Client
from mcipc.rcon.functions import parsed
from mcipc.rcon.response_types.difficulty import parse
from mcipc.rcon.types import Difficulty


__all__ = ['difficulty']


@parsed(parse)
def difficulty(self: Client, difficulty: Difficulty) -> bool:
    """Sets the difficulty."""

    return self.run('difficulty', difficulty)



"""Implementation of the replaceitem command."""

from typing import Optional

from mcipc.rcon.be.types import EntityEquipmentSlot, ReplaceMode
from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import Vec3


__all__ = ['ReplaceitemProxy', 'replaceitem']


class ReplaceitemProxy(CommandProxy):
    """Proxy for replaceitem related commands."""

    # pylint: disable=R0913
    def block(
            self,
            position: Vec3,
            slot_id: int,
            item_name: str,
            amount: Optional[int] = None,
            data: Optional[int] = None,
            components: Optional[dict] = None,
            *,
            old_item_handling: Optional[ReplaceMode] = None
    ) -> str:
        """Replaces a block."""
        command = ['block', position, 'slot.container', slot_id]

        if old_item_handling is not None:
            command.append(old_item_handling)

        return self._run(*command, item_name, amount, data, components)

    def entity(
            self,
            target: str,
            slot_type: EntityEquipmentSlot,
            slot_id: int,
            item_name: str,
            amount: Optional[int] = None,
            data: Optional[int] = None,
            components: Optional[dict] = None,
            *,
            old_item_handling: Optional[ReplaceMode] = None
    ) -> str:
        """Replaces an item."""
        command = ['entity', target, slot_type, slot_id]

        if old_item_handling is not None:
            command.append(old_item_handling)

        return self._run(*command, item_name, amount, data, components)


def replaceitem(self: Client) -> ReplaceitemProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.be.commands.replaceitem.ReplaceitemProxy`
    """

    return ReplaceitemProxy(self, 'replaceitem')



"""Implementation of the effect command."""

from typing import Optional

from mcipc.rcon.client import Client


__all__ = ['effect']


def effect(
        self: Client,
        player: str,
        effect_or_clear: str,
        seconds: Optional[int] = None,
        amplifier: Optional[int] = None,
        hide_particles: Optional[bool] = None
) -> str:
    """Returns a proxy for available sub-commands."""

    command = ['effect', player, (mode := str(effect_or_clear))]

    if mode != 'clear':
        command += [seconds, amplifier, hide_particles]

    return self.run(*command)



"""Implementation of the testforblock command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.types import Vec3


__all__ = ['testforblock']


def testforblock(
        self: Client,
        position: Vec3,
        tile_name: str,
        data_value: Optional[int] = None
) -> str:
    """Tests whether a certain block is in a specific location."""

    return self.run('testforblock', position, tile_name, data_value)



"""Implements the toggledownfall command."""

from mcipc.rcon.client import Client


__all__ = ['toggledownfall']


def toggledownfall(self: Client) -> str:
    """Toggles downfall."""

    return self.run('toggledownfall')



"""Operators management."""

from mcipc.rcon.client import Client


__all__ = ['deop', 'op']


def deop(self: Client, player: str) -> str:
    """Revokes operator status from the respective player."""

    return self.run('deop', player)


def op(self: Client, player: str) -> str:     # pylint: disable=C0103
    """Makes the respective player an operator."""

    return self.run('op', player)



"""Implementation of the structure command."""

from typing import Optional

from mcipc.rcon.be.types import Mirror
from mcipc.rcon.be.types import Rotation
from mcipc.rcon.be.types import StructureAnimationMode
from mcipc.rcon.be.types import StructureSaveMode
from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import Vec3


__all__ = ['StructureProxy', 'structure']


class StructureProxy(CommandProxy):
    """Proxy for structure commands."""

    def save(
            self,
            name: str,
            from_: Vec3,
            to: Vec3,
            save_mode: Optional[StructureSaveMode] = None,
            includes_blocks: Optional[bool] = None,
            *,
            includes_entities: Optional[bool] = None
    ) -> str:
        """Saves a structure."""

        command = ['save', name, from_, to]

        if includes_entities is not None:
            command.append(includes_entities)

        return self._run(*command, save_mode, includes_blocks)

    def load(
            self,
            name: str,
            to: Vec3,
            rotation: Optional[Rotation] = None,
            mirror: Optional[Mirror] = None,
            includes_entities: Optional[bool] = None,
            includes_blocks: Optional[bool] = None,
            integrity: Optional[float] = None,
            seed: str = Optional[None],
            *,
            animation_mode: Optional[StructureAnimationMode] = None,
            animation_seconds: Optional[float] = None
    ) -> str:
        """Loads a structure."""
        command = ['load', name, to, rotation, mirror]

        if animation_mode is not None and animation_seconds is not None:
            command += [animation_mode, animation_seconds]

        return self._run(*command, includes_entities, includes_blocks,
                         integrity, seed)


def structure(self: Client) -> StructureProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.be.commands.structure.StructureProxy`
    """

    return StructureProxy(self, 'structure')



"""Implementation of the defaultgamemode command."""

from mcipc.rcon.client import Client
from mcipc.rcon.types import GameMode


__all__ = ['defaultgamemode']


def defaultgamemode(self: Client, mode: GameMode) -> str:
    """Sets the default game mode."""

    return self.run('defaultgamemode', mode)



"""Implementation of the reload command."""

from mcipc.rcon.client import Client


__all__ = ['reload']


def reload(self: Client) -> str:
    """Reloads data packs in Java Edition and functions
    in behavior packs in Bedrock Edition.
    """

    return self.run('reload')



"""Full statistics protocol."""

from __future__ import annotations
from typing import IO, Iterator, NamedTuple

from mcipc.functions import json_serializable
from mcipc.query.proto.common import MAGIC
from mcipc.query.proto.common import NULL
from mcipc.query.proto.common import ip_or_hostname
from mcipc.query.proto.common import random_session_id
from mcipc.query.proto.common import BigEndianSignedInt32
from mcipc.query.proto.common import IPAddressOrHostname
from mcipc.query.proto.common import Type


__all__ = ['Request', 'FullStats']


PADDING = b'\x00\x00\x00\x00'


def read_stats(file: IO) -> dict:
    """Returns the end index and a dictionary
    of zero-separated key-value pairs.
    """

    item = ''
    dictionary = {}
    key = None
    is_key = True

    while True:
        byte = file.read(1)

        if byte == NULL:
            if not item and is_key:
                return dictionary

            if is_key:
                key = item
                is_key = False
            else:
                dictionary[key] = item
                key = None
                is_key = True

            item = ''
        else:
            item += byte.decode('latin-1')


def read_players(file: IO) -> Iterator[str]:
    """Yields zero-byte-separated items."""

    item = ''

    while True:
        byte = file.read(1)

        if byte == NULL:
            if not item:
                return

            yield item
            item = ''
        else:
            item += byte.decode('latin-1')


def plugins_to_dict(string: str) -> dict:
    """Convers a plugins string into a dictionary."""

    try:
        mod, plugins = string.split(': ')
    except ValueError:  # No plugins.
        return {}

    return {mod: plugins.split('; ')}


def stats_from_dict(dictionary: dict):
    """Yields statistics options from the provided dictionary."""

    yield dictionary['hostname']
    yield dictionary['gametype']
    yield dictionary['game_id']
    yield dictionary['version']
    yield plugins_to_dict(dictionary['plugins'])
    yield dictionary['map']
    yield int(dictionary['numplayers'])
    yield int(dictionary['maxplayers'])
    yield int(dictionary['hostport'])
    yield ip_or_hostname(dictionary['hostip'])


class Request(NamedTuple):
    """Basic statistics request packet."""

    magic: bytes = MAGIC
    type: Type = Type.STAT
    session_id: BigEndianSignedInt32 = BigEndianSignedInt32()
    challenge_token: BigEndianSignedInt32 = BigEndianSignedInt32()
    padding: bytes = PADDING

    def __bytes__(self):
        """Returns the packet as bytes."""
        payload = self.magic
        payload += bytes(self.type)
        payload += bytes(self.session_id)
        payload += bytes(self.challenge_token)
        payload += self.padding
        return payload

    @classmethod
    def create(cls, challenge_token: BigEndianSignedInt32) -> Request:
        """Creates a new request with the given challenge token."""
        return cls(session_id=random_session_id(),
                   challenge_token=challenge_token)


@json_serializable
class FullStats(NamedTuple):
    """Full statistics response."""

    type: Type
    session_id: BigEndianSignedInt32
    host_name: str
    game_type: str
    game_id: str
    version: str
    plugins: dict
    map: str
    num_players: int
    max_players: int
    host_port: int
    host_ip: IPAddressOrHostname
    players: list[str]

    @classmethod
    def read(cls, file: IO) -> FullStats:
        """Read a full stats response."""
        type_ = Type.read(file)
        session_id = BigEndianSignedInt32.read(file)
        file.read(11)   # Discard padding.
        stats = read_stats(file)
        file.read(10)   # Discard padding.
        players = list(read_players(file))
        return cls(type_, session_id, *stats_from_dict(stats), players)



"""Implementation of the schedule command."""

from pathlib import Path

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import Vec3


__all__ = ['ScheduleProxy', 'schedule']


class ScheduleProxy(CommandProxy):
    """Proxy for schedule-related commands."""

    # pylint: disable=C0103
    def __call__(self, from_: Vec3, to: Vec3, function: Path) -> str:
        """Adds a schedule."""
        return self._run(from_, to, function)

    def circle(self, center: Vec3, radius: int, function: Path) -> str:
        """Adds a circle."""
        return self._run('circle', center, radius, function)

    def tickingarea(self, name: str, function: Path) -> str:
        """Adds a ticking ares."""
        return self._run('tickingarea', name, function)


def schedule(self: Client) -> ScheduleProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.be.commands.schedule.ScheduleProxy`
    """

    return ScheduleProxy(self, 'schedule', 'on_area_loaded', 'add')



"""Query client library."""

from socket import SOCK_DGRAM, socket
from typing import Optional, Union
from warnings import warn

from mcipc.query.proto import BasicStats
from mcipc.query.proto import BasicStatsRequest
from mcipc.query.proto import BigEndianSignedInt32
from mcipc.query.proto import FullStats
from mcipc.query.proto import FullStatsRequest
from mcipc.query.proto import HandshakeRequest
from mcipc.query.proto import Response


__all__ = ['Client']


WARN_TEMP = 'Client.{} is deprecated. Use Client.stats({}) instead.'
BasicMessages = tuple[BasicStatsRequest, BasicStats]
FullMessages = tuple[FullStatsRequest, FullStats]


def get_message_types(full: bool) -> Union[BasicMessages, FullMessages]:
    """Returns request and response types."""

    if full:
        return (FullStatsRequest, FullStats)

    return (BasicStatsRequest, BasicStats)


class Client:
    """A basic client, common to Query and RCON."""

    def __init__(self, host: str, port: int, *, timeout: float = None):
        """Sets host an port."""
        self._socket = socket(type=SOCK_DGRAM)
        self.host = host
        self.port = port
        self.timeout = timeout
        self.challenge_token = None

    def connect(self) -> None:
        """Contects the socket."""
        self._socket.__enter__()
        self._socket.connect((self.host, self.port))

        if self.challenge_token is None:
            self.challenge_token = self.handshake()

    def disconnect(self) -> Optional[bool]:
        """Delegates to the underlying socket's exit method."""
        self.challenge_token = None
        return self._socket.__exit__()

    def __enter__(self):
        """connect on entering a context"""
        self.connect()
        return self

    def __exit__(self, *_):
        """Delegates to the underlying socket's exit method."""
        return self.disconnect()

    @property
    def timeout(self) -> float:
        """Returns the socket timeout."""
        return self._socket.gettimeout()

    @timeout.setter
    def timeout(self, timeout: float):
        """Sets the socket timeout."""
        self._socket.settimeout(timeout)

    @property
    def basic_stats(self) -> BasicStats:
        """Returns basic stats."""
        warn(WARN_TEMP.format('basic_stats', ''), FutureWarning, stacklevel=2)
        return self.stats()

    @property
    def full_stats(self) -> FullStats:
        """Returns full stats."""
        warn(WARN_TEMP.format('full_stats', 'full=True'), FutureWarning,
             stacklevel=2)
        return self.stats(full=True)

    def handshake(self) -> BigEndianSignedInt32:
        """Performs a handshake."""
        request = HandshakeRequest.create()

        with self._socket.makefile('wb') as file:
            file.write(bytes(request))

        with self._socket.makefile('rb') as file:
            response = Response.read(file)

        return response.challenge_token

    def stats(self, full: bool = False) -> Union[BasicStats, FullStats]:
        """Returns basic or full stats."""
        request_type, return_type = get_message_types(full)
        request = request_type.create(self.challenge_token)

        with self._socket.makefile('wb') as file:
            file.write(bytes(request))

        with self._socket.makefile('rb') as file:
            return return_type.read(file)



"""Implementation of the effect command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy


__all__ = ['EffectProxy', 'effect']


class EffectProxy(CommandProxy):
    """Proxy for effect commands on the Java Edition."""

    # pylint: disable=W0621
    def clear(
            self,
            targets: Optional[str] = None,
            effect: Optional[str] = None
    ) -> str:
        """Clears effects."""
        return self._run('clear', targets, effect)

    def give(
            self,
            targets: str,
            effect: str,
            seconds: Optional[int] = None,
            amplifier: Optional[int] = None,
            hide_particles: Optional[bool] = None
    ) -> str:
        """Gives an effect."""
        return self._run('give', targets, effect, seconds, amplifier,
                         hide_particles)


def effect(self: Client) -> EffectProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.effect.EffectProxy`
    """

    return EffectProxy(self, 'effect')



"""Implementation of the datapack command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import DatapackMode, DatapackState


__all__ = ['DatapackProxy', 'datapack']


class DatapackProxy(CommandProxy):
    """Proxy for datapack related commands."""

    def disable(self, name: str) -> str:
        """Disables a datapack."""
        return self._run('disable', name)

    def enable(
            self,
            name: str,
            mode: DatapackMode,
            existing: Optional[str] = None
    ) -> str:
        """Enables a datapack."""
        command = ['enable', name, mode]

        if mode in {DatapackMode.AFTER, DatapackMode.BEFORE}:
            if existing is None:
                raise ValueError('Missing value for existing datapack.')

            command.append(existing)

        return self._run(*command)

    def list(self, state: Optional[DatapackState] = None) -> str:
        """Lists the enabled datapacks."""
        command = ['list']

        if state is not None:
            command.append(state)

        return self._run(*command)


def datapack(self: Client) -> DatapackProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.datapack.DatapackProxy`
    """

    return DatapackProxy(self, 'datapack')



"""Implementation of the experience and xp commands."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.proxy import CommandProxy
from mcipc.rcon.types import XPUnit


__all__ = ['ExperienceProxy', 'experience', 'xp']


class ExperienceProxy(CommandProxy):
    """Proxy for Experience commands."""

    def add(
            self,
            targets: str,
            amount: int,
            unit: Optional[XPUnit] = None
    ) -> str:
        """Gives the target the specified amount of XP units."""
        return self._run('add', targets, amount, unit)

    def set(
            self,
            targets: str,
            amount: int,
            unit: Optional[XPUnit] = None
    ) -> str:
        """Sets the XP of the target to the specified amount and units."""
        return self._run('set', targets, amount, unit)

    def query(self, targets: str, amount: int, unit: XPUnit) -> str:
        """Queries the amount in given unit of XP of the targets."""
        return self._run('query', targets, amount, unit)


def experience(self: Client) -> ExperienceProxy:
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.experience.ExperienceProxy`
    """

    return ExperienceProxy(self, 'experience')


def xp(self: Client) -> ExperienceProxy:    # pylint: disable=C0103
    """Delegates to a
    :py:class:`mcipc.rcon.je.commands.experience.ExperienceProxy`
    """

    return ExperienceProxy(self, 'xp')



"""The actual server."""

from json import dumps
from logging import getLogger
from socket import socket
from typing import IO

from mcipc.server.datastructures import Handshake, SLPResponse
from mcipc.server.datatypes import VarInt
from mcipc.server.enumerations import State


__all__ = ['MAX_PLAYERS', 'PROTOCOL', 'VERSION', 'get_response', 'StubServer']


LOGGER = getLogger(__file__)
MAX_PLAYERS = 20
PROTOCOL = 753
VERSION = '1.18.1'


def get_response(text: str) -> bytes:
    """Returns the response text message."""

    payload = dumps({'text': text}).encode('latin-1')
    payload_length = bytes(VarInt(len(payload)))
    payload = bytes(VarInt(0)) + payload_length + payload
    payload_length = bytes(VarInt(len(payload)))
    return payload_length + payload


class StubServer:
    """A stub minecraft server."""

    def __init__(self, description: str, *, version: str = VERSION,
                 max_players: int = MAX_PLAYERS, protocol: int = PROTOCOL):
        """Description, max players and protocol information."""
        self.description = description
        self.version = version
        self.max_players = max_players
        self.protocol = protocol

    @property
    def slp_content(self) -> dict:
        """Returns the content of an SLP response."""
        return {
            'version': {
                'name': self.version,
                'protocol': self.protocol
            },
            'players': {
                'max': self.max_players,
                'online': 0
            },
            'description': {
                'text': self.description
            }
        }

    @property
    def slp_response(self) -> SLPResponse:
        """Returns an SLP response."""
        return SLPResponse(VarInt(0), self.slp_content)

    @staticmethod
    def _perform_handshake(rfile: IO) -> State:
        """Handle handshake requests."""
        handshake = Handshake.read(rfile)
        LOGGER.debug('Got handshake: %s', handshake)
        return handshake.next_state

    def _perform_status(self, rfile: IO, wfile: IO):
        """Handles status requests."""
        packet_id = rfile.read(1)

        if packet_id == b'\x01':
            LOGGER.debug('Got packet id: %s', packet_id)
            wfile.write(bytes(self.slp_response))

    def _perform_login(self, wfile: IO):
        """Handles the login response."""
        raise NotImplementedError()

    def _handle_login(self, rfile: IO, wfile: IO):
        """Performs a login."""
        size = VarInt.read(rfile)
        packet_id = VarInt.read(rfile)
        packet_id_length = len(bytes(packet_id))
        payload = rfile.read(size - packet_id_length)
        LOGGER.debug('Got packet ID: %s', packet_id)
        user_name = payload[2:].decode('latin-1')
        LOGGER.debug('User "%s" logged in.', user_name)
        self._perform_login(wfile)

    def _process(self, rfile: IO, wfile: IO, state: State = State.HANDSHAKE):
        """Runs the connection processing."""
        LOGGER.debug('Current state: %s', state)

        if state == State.HANDSHAKE:
            state = self._perform_handshake(rfile)
            self._process(rfile, wfile, state=state)
        elif state == State.STATUS:
            self._perform_status(rfile, wfile)
        elif state == State.LOGIN:
            self._handle_login(rfile, wfile)

    def _main_loop(self, sock: socket):
        """Runs the main server loop."""

        conn, addr = sock.accept()

        with conn.makefile('rb') as rfile, conn.makefile('wb') as wfile:
            LOGGER.debug('New connection from: %s', addr)
            self._process(rfile, wfile)

    def spawn(self, address: str, port: int):
        """Spawns the server on the respective socket."""
        with socket() as sock:
            sock.bind((address, port))
            sock.listen()

            while True:
                self._main_loop(sock)



"""Client for the education edition."""

from mcipc.rcon.be.commands.ability import ability
from mcipc.rcon.be.commands.clear import clear
from mcipc.rcon.be.commands.clone import clone
from mcipc.rcon.be.commands.difficulty import difficulty
from mcipc.rcon.be.commands.effect import effect
from mcipc.rcon.be.commands.enchant import enchant
from mcipc.rcon.be.commands.fill import fill
from mcipc.rcon.be.commands.gamemode import gamemode
from mcipc.rcon.be.commands.give import give
from mcipc.rcon.be.commands.help import help    # pylint: disable=W0622
from mcipc.rcon.be.commands.list import list    # pylint: disable=W0622
from mcipc.rcon.be.commands.locate import locate
from mcipc.rcon.be.commands.mobevent import mobevent
from mcipc.rcon.be.commands.playsound import playsound
from mcipc.rcon.be.commands.replaceitem import replaceitem
from mcipc.rcon.be.commands.setblock import setblock
from mcipc.rcon.be.commands.setmaxplayers import setmaxplayers
from mcipc.rcon.be.commands.setworldspawn import setworldspawn
from mcipc.rcon.be.commands.spawnpoint import spawnpoint
from mcipc.rcon.be.commands.spreadplayers import spreadplayers
from mcipc.rcon.be.commands.stopsound import stopsound
from mcipc.rcon.be.commands.summon import summon
from mcipc.rcon.be.commands.teleport import teleport
from mcipc.rcon.be.commands.testfor import testfor
from mcipc.rcon.be.commands.testforblock import testforblock
from mcipc.rcon.be.commands.testforblocks import testforblocks
from mcipc.rcon.be.commands.tickingarea import tickingarea
from mcipc.rcon.be.commands.time import time
from mcipc.rcon.be.commands.title import title
from mcipc.rcon.be.commands.toggledownfall import toggledownfall
from mcipc.rcon.be.commands.transferserver import transferserver
from mcipc.rcon.be.commands.weather import weather
from mcipc.rcon.be.commands.worldbuilder import wb, worldbuilder
from mcipc.rcon.be.commands.wsserver import connect, wsserver
from mcipc.rcon.be.commands.xp import xp
from mcipc.rcon.ee.commands.agent import collect
from mcipc.rcon.ee.commands.agent import createagent
from mcipc.rcon.ee.commands.agent import destroy
from mcipc.rcon.ee.commands.agent import detect
from mcipc.rcon.ee.commands.agent import detectredstone
from mcipc.rcon.ee.commands.agent import dropall
from mcipc.rcon.ee.commands.agent import move
from mcipc.rcon.ee.commands.agent import remove
from mcipc.rcon.ee.commands.agent import tpagent
from mcipc.rcon.ee.commands.agent import transfer
from mcipc.rcon.ee.commands.agent import turn
from mcipc.rcon.ee.commands.classroommode import classroommode
from mcipc.rcon.ee.commands.code import code
from mcipc.rcon.ee.commands.geteduclientinfo import geteduclientinfo
from mcipc.rcon.ee.commands.immutableworld import immutableworld
from mcipc.rcon.ee.commands.position import position
from mcipc.rcon.commands.chat import me, say, tell
from mcipc.rcon.commands.execute import execute
from mcipc.rcon.commands.gamerule import gamerule
from mcipc.rcon.commands.kick import kick
from mcipc.rcon.commands.kill import kill
from mcipc.rcon.commands.op import deop, op
from mcipc.rcon.client import Client as _Client


__all__ = ['Client']


class Client(_Client):
    """Client for the Education Edition."""

    ability = ability
    classroommode = classroommode
    clear = clear
    clone = clone
    code = code
    collect = collect
    connect = connect
    createagent = createagent
    deop = deop
    destroy = destroy
    detect = detect
    detectredstone = detectredstone
    difficulty = difficulty
    dropall = dropall
    effect = effect
    enchant = enchant
    execute = property(execute)
    fill = fill
    gamemode = gamemode
    gamerule = gamerule
    geteduclientinfo = geteduclientinfo
    give = give
    help = help
    immutableworld = immutableworld
    kick = kick
    kill = kill
    list = list
    locate = locate
    me = me
    mobevent = mobevent
    move = move
    op = op
    playsound = playsound
    position = position
    remove = remove
    replaceitem = property(replaceitem)
    say = say
    setblock = setblock
    setmaxplayers = setmaxplayers
    setworldspawn = setworldspawn
    spawnpoint = spawnpoint
    spreadplayers = spreadplayers
    stopsound = stopsound
    summon = summon
    teleport = teleport
    tell = msg = w = tell
    testfor = testfor
    testforblock = testforblock
    testforblocks = testforblocks
    tickingarea = property(tickingarea)
    time = property(time)
    title = title
    toggledownfall = toggledownfall
    tpagent = tpagent
    transfer = transfer
    transferserver = transferserver
    turn = turn
    wb = wb
    weather = property(weather)
    worldbuilder = worldbuilder
    wsserver = wsserver
    xp = experience = xp



"""Implementation of the save command."""

from mcipc.rcon.client import Client
from mcipc.rcon.be.types import SaveCommand


__all__ = ['save']


def save(self: Client, command: SaveCommand) -> str:
    """Issues the specified save command."""

    return self.run('save', command)



"""Implementation of the clearspawnpoint."""

from typing import Optional

from mcipc.rcon.client import Client


__all__ = ['clearspawnpoint']


def clearspawnpoint(self: Client, player: Optional[str] = None) -> str:
    """Used to remove spawnpoints in the world."""

    return self.run('clearspawnpoint', player)



"""Implementation of the testfor command."""

from mcipc.rcon.client import Client


__all__ = ['testfor']


def testfor(self: Client, victim: str) -> str:
    """Counts entities matching specified conditions."""

    return self.run('testfor', victim)



"""Implementation of the spawnpoint command."""

from typing import Optional

from mcipc.rcon.client import Client
from mcipc.rcon.types import Vec3


__all__ = ['spawnpoint']


def spawnpoint(
        self: Client,
        player: Optional[str] = None,
        spawn_pos: Optional[Vec3] = None
) -> str:
    """Sets the spawn point for a player."""

    return self.run('spawnpoint', player, spawn_pos)



"""Parsing responses from the difficulty command."""

from mcipc.rcon.functions import parse_bool


__all__ = ['parse']


SET = r'The difficulty has been set to (\w+)'
UNCHANGED = r'The difficulty did not change; it is already set to (\w+)'


def parse(text: str) -> bool:
    """Parses a boolean value from the text
    returned by the difficulty command.
    """

    return parse_bool(text, true=SET, false=UNCHANGED)



#! /usr/bin/env python
"""Installation script."""

from setuptools import setup

with open('README.md') as file:
    long_description = file.read()

setup(
    name='mcipc',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author='Richard Neumann',
    author_email='mail@richard-neumann.de',
    python_requires='>=3.9',
    packages=[
        'mcipc',
        'mcipc.query',
        'mcipc.query.proto',
        'mcipc.rcon',
        'mcipc.rcon.be',
        'mcipc.rcon.be.commands',
        'mcipc.rcon.commands',
        'mcipc.rcon.ee',
        'mcipc.rcon.ee.commands',
        'mcipc.rcon.je',
        'mcipc.rcon.je.commands',
        'mcipc.rcon.response_types',
        'mcipc.server'
    ],
    install_requires=['rcon'],
    entry_points={
        'console_scripts': [
            'mcstubsrv = mcipc.server.stubsrv:main',
            'queryclt = mcipc.query.queryclt:main'
        ],
    },
    url='https://github.com/conqp/mcipc',
    license='GPLv3',
    description='A Minecraft server inter-process communication library.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='minecraft python server rcon query'
)



"""Common protocol stuff."""

from __future__ import annotations
from enum import Enum
from functools import partial
from ipaddress import IPv4Address, IPv6Address, ip_address
from random import randint
from typing import IO, Iterable, Iterator, Union


__all__ = [
    'MAGIC',
    'NULL',
    'decode_all',
    'ip_or_hostname',
    'random_session_id',
    'BigEndianSignedInt32',
    'IPAddressOrHostname',
    'Type'
]


MAGIC = b'\xfe\xfd'
NULL = b'\0'
SESSION_ID_MASK = 0x0F0F0F0F


IPAddressOrHostname = Union[IPv4Address, IPv6Address, str]


def decode_all(blocks: Iterable[bytes], encoding='latin-1') -> Iterator[str]:
    """Decodes all byte blocks with the given encoding."""

    return map(partial(bytes.decode, encoding=encoding), blocks)


def ip_or_hostname(string: str) -> IPAddressOrHostname:
    """Returns an IPv4 or IPv6 address if applicable, else a string."""

    try:
        return ip_address(string)
    except ValueError:
        return string


def random_session_id() -> BigEndianSignedInt32:
    """Returns a random session ID.
    See: https://wiki.vg/Query#Generating_a_Session_ID
    """

    random = randint(BigEndianSignedInt32.MIN, BigEndianSignedInt32.MAX)
    return BigEndianSignedInt32(random & SESSION_ID_MASK)


class BigEndianSignedInt32(int):
    """A big-endian, signed int32."""

    MIN = -2_147_483_648
    MAX = 2_147_483_647

    def __init__(self, *_):
        """Checks the boundaries."""
        super().__init__()

        if not self.MIN <= self <= self.MAX:
            raise ValueError('Signed int32 out of bounds:', self)

    def __bytes__(self):
        """Returns the int as bytes."""
        return self.to_bytes(4, 'big', signed=True)

    @classmethod
    def from_bytes(cls, bytes_: bytes) -> BigEndianSignedInt32:
        """Returns the int from the given bytes."""
        return cls(super().from_bytes(bytes_, 'big', signed=True))

    @classmethod
    def read(cls, file: IO) -> BigEndianSignedInt32:
        """Reads a big endian int32 from a file-like object."""
        return cls.from_bytes(file.read(4))


class Type(int, Enum):
    """Request type."""

    HANDSHAKE = 9
    STAT = 0

    def __bytes__(self):
        """Returns the integer value as big endian."""
        return self.value.to_bytes(1, 'big')

    @classmethod
    def from_bytes(cls, bytes_: bytes) -> Type:
        """Returns the type from the respective bytes."""
        return cls(int.from_bytes(bytes_, 'big'))

    @classmethod
    def read(cls, file: IO) -> Type:
        """Reads the type from a file-like object."""
        return cls.from_bytes(file.read(1))



"""Parses a seed value from a server response."""

from re import fullmatch


__all__ = ['parse']


REGEX = r'.*\[(-?\d+)\]'


def parse(text: str) -> int:
    """Returns an integer."""

    if (match := fullmatch(REGEX, text)) is None:
        raise ValueError('Unexpected seed response:', text)

    return int(match.group(1))



"""Implementation of the difficulty command."""

from typing import Union

from mcipc.rcon.client import Client
from mcipc.rcon.types import Difficulty


__all__ = ['difficulty']


# pylint: disable=W0621
def difficulty(self: Client, difficulty: Union[Difficulty, int]) -> str:
    """Sets the difficulty."""

    return self.run('difficulty', difficulty)



"""Implementation of the classroommode command."""

from mcipc.rcon.client import Client


__all__ = ['classroommode']


def classroommode(self: Client):
    """Unknown invocation and behaviour.
    See: https://minecraft.gamepedia.com/Commands/classroommode
    """

    raise NotImplementedError(__doc__)



"""Query server configuration."""

from __future__ import annotations
from configparser import ConfigParser, SectionProxy
from os import getenv, name
from pathlib import Path
from typing import Iterator, NamedTuple

from mcipc.query.exceptions import InvalidConfig


__all__ = ['CONFIG']


CONFIG = ConfigParser()

if name == 'posix':
    CONFIG_FILE = Path('/etc/query.conf')
elif name == 'nt':
    CONFIG_FILE = Path(getenv('LOCALAPPDATA')) / 'query.conf'
else:
    raise NotImplementedError(f'Unsupported operating system: {name}')


class Config(NamedTuple):
    """Represents server configuration."""

    host: str
    port: int

    @classmethod
    def from_string(cls, string: str) -> Config:
        """Reads the credentials from the given string."""
        try:
            host, port = string.split(':')
        except ValueError:
            raise InvalidConfig(f'Invalid socket: {string}.') from None

        try:
            port = int(port)
        except ValueError:
            raise InvalidConfig(f'Not an integer: {port}.') from None

        return cls(host, port)

    @classmethod
    def from_config_section(cls, section: SectionProxy) -> Config:
        """Creates a credentials tuple from
        the respective config section.
        """
        host = section['host']
        port = section.getint('port')
        return cls(host, port)


def entries(config_parser: ConfigParser) -> Iterator[tuple[str, Config]]:
    """Yields config entries."""

    for section in config_parser.sections():
        yield (section, Config.from_config_section(config_parser[section]))


def servers() -> dict[str, Config]:
    """Returns a dictionary of servers."""

    CONFIG.read(CONFIG_FILE)
    return dict(entries(CONFIG))
